import bpy
import os
import sys
from pathlib import Path
from math import degrees
import bmesh


# print('starteddddddddddddddddddd')
# for i, arg in enumerate(sys.argv):
#     print('aaaaaaaaaaarg:', arg)

blender_executable, blend_file = sys.argv[:2]
script_file = sys.argv[4]
out_file_path = sys.argv[5]
decimals = int([e for e in sys.argv if e.startswith('--decimals=')][0][len('--decimals='):])
TRIANGULATE = '--triangulate' in sys.argv


print('-----------------------', 'decimals:', decimals, 'triangulate:', TRIANGULATE)

scene_name = Path(bpy.data.filepath).stem
print('file_name:', scene_name)

code = ''

if not '--models_only' in sys.argv:
    code += '''from ursina import *
from time import perf_counter

scene_parent = Entity()

if __name__ == '__main__':
    app = Ursina()

t = perf_counter()
'''
code += '''
# unique meshes
meshes = {
'''

dg = bpy.context.evaluated_depsgraph_get() #getting the dependency graph
bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
objects = [ob for ob in bpy.data.objects if ob.type == 'MESH']
unique_objects = {}
# print('meshes:', meshes)
for ob in objects:
    unique_objects[ob.data.name] = ob


for key, ob in unique_objects.items():
    polygons =[]
    verts = []
    normals = []
    vertex_colors = []
    uvs = []
    indices = []

    mesh = ob.evaluated_get(dg).data
    # for poly in mesh.polygons:
    #     print('---------------poly:', [int(e) for e in poly.vertices])
    if TRIANGULATE:
        # triangulate mesh
        bm = bmesh.new()
        bm.from_mesh(mesh)
        bmesh.ops.triangulate(bm, faces=bm.faces[:], quad_method='BEAUTY', ngon_method='BEAUTY')
        bm.normal_update()
        bm.to_mesh(mesh)
        bm.free()
    else:
        for poly in mesh.polygons:
            # print('---------------poly:', [int(e) for e in poly.vertices])
            polygons.append([int(e) for e in poly.vertices])


    for poly in mesh.polygons:
        indices.extend(poly.vertices)

    if TRIANGULATE:
        for idx in indices:
            v = mesh.vertices[idx]
            verts.append((round(v.co[0],decimals),round(v.co[2],decimals),round(v.co[1],decimals)))
    else:
        verts = [(round(v.co[0],decimals),round(v.co[2],decimals),round(v.co[1],decimals)) for v in mesh.vertices]

    if '--vertex_colors' in sys.argv and 'Color' in mesh.attributes:
        color_data = mesh.attributes['Color'].data
        vertex_colors_data = [[round(e,4) for e in (color_data[vertex.index].color_srgb)] for vertex in mesh.vertices]
        vertex_colors = [vertex_colors_data[idx] for idx in indices]

    if '--normals' in sys.argv:
        for idx in indices:
            v = mesh.vertices[idx]
            normals.append((v.normal[0], v.normal[2], v.normal[1]))
    # average the vertex normals to get face normals
        sharp_normals = []
        for i in range(0, len(normals), 3):
            averaged_normal = (
                round(sum((normals[i][0], normals[i+1][0], normals[i+2][0])) / 3, decimals),
                round(sum((normals[i][1], normals[i+1][1], normals[i+2][1])) / 3, decimals),
                round(sum((normals[i][2], normals[i+1][2], normals[i+2][2])) / 3, decimals),
                )
            for j in range(3):
                sharp_normals.append(averaged_normal)

        normals = sharp_normals


    if '--uvs' in sys.argv:
        uv_layer = mesh.uv_layers.active
        if uv_layer and len(uv_layer.data) > 0:
            uvs = [tuple(round(e, decimals) for e in data.uv) for data in uv_layer.data]



    code += f'''
'{mesh.name}' : Mesh(
    vertices={str(verts).replace(' ', '')},
    triangles={polygons},
    normals={str(normals).replace(' ', '')},
    colors={str(vertex_colors).replace(' ', '')},
    uvs={str(uvs).replace(' ', '')},
),
'''

code += '}\n'

if '--models_only' in sys.argv:
    print('sucessfully exported blender models')
    with open(out_file_path, 'w') as f:
        f.write(code)

else:
    # write entities
    code += '''print('loaded models:', perf_counter() - t)\n'''
    code += '''t = perf_counter()\n'''


    for ob in objects:
        if '--skip_hidden' in sys.argv and ob.hide_get() == True:
            continue

        #deselect all but just one object and make it active
        bpy.ops.object.select_all(action='DESELECT')
        rotation = [degrees(e) for e in ob.matrix_world.to_euler()]

        code += f'''
scene_parent.{ob.name.replace('.', '_')} = Entity(
    name=\'{ob.name}\',
    parent=scene_parent,
    position=Vec3({round(ob.location.x,5)}, {round(ob.location.z,5)}, {round(ob.location.y,5)}),
    rotation=({round(-rotation[0],5)}, {round(-rotation[2],5)}, {round(rotation[1],5)}),
    scale=Vec3({round(ob.scale.x,5)}, {round(ob.scale.z,5)}, {round(ob.scale.y,5)}),
    '''

        code += f'''model=copy(meshes['{ob.data.name}']),'''
        # code += f'''model='cube','''

        if ob.active_material:
            color = ob.active_material.diffuse_color
            code += f'''
    color=({round(color[0],5)}, {round(color[1],5)}, {round(color[2],5)}, {round(color[3],5)}),'''

        code += f'''
    ignore=True,
    )'''
    code += '''\n\nscene_parent.meshes = meshes'''
    code += '''\nprint('created entities:', perf_counter() - t)\n'''

    code += '''
if __name__ == '__main__':
    EditorCamera()
    app.run()
'''

    # return code
    print('sucessfully exported blender scene')
    with open(out_file_path, 'w') as f:
        f.write(code)



from ursina import Shader, color

noise_fog_shader = Shader(name='noise_fog_shader', language=Shader.GLSL, fragment='''
#version 130

uniform vec4 p3d_ColorScale;
in vec2 uv;
out vec4 result;

uniform sampler2D p3d_Texture0;
uniform float time;

uniform vec4 dark_color;
uniform vec4 light_color;

void main() {
    vec4 a = texture(p3d_Texture0, uv + vec2(time*.1, time*.1));
    vec4 b = texture(p3d_Texture0, uv - vec2(time*.1, time*.075));
    vec4 c = (a + b) * .5;

    result = texture(p3d_Texture0, uv + c.xy) * c;
    result.a = 1.0;

    result = mix(dark_color, light_color, result.r);
}

''',  # noqa: I001
default_input = {
  'dark_color' : color.black,
  'light_color' : color.white,
}
)



if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    window.color=color.black

    e = Entity(model='quad', texture='perlin_noise', shader=noise_fog_shader, scale=6)
    e.set_shader_input('dark_color', hsv(280,1,.1,0))
    e.set_shader_input('light_color', color.cyan)

    app.t = 0
    def update():
        app.t += time.dt
        e.set_shader_input('time', app.t)

    bg = Sprite('shore', scale=1, z=1, color=color.dark_gray)
    window.size *= .5
    app.run()



from panda3d.core import Vec4 as PandaVec4


class Vec4(PandaVec4):

    def __round__(self, decimals=4):
        return Vec4(*(round(e,decimals) for e in self))


    def __repr__(self):
        return f'Vec4({self[0]}, {self[1]}, {self[2]}, {self[3]})'


    def __iadd__(self, value):
        if len(value) % 3 == 0:
            for i in range(0, len(value), 3):
                self.add_x(value[i])
                self.add_y(value[i+1])
                self.add_z(value[i+2])
                return self

        if len(value) % 2 == 0:
            for i in range(0, len(value), 2):
                self.add_x(value[i])
                self.add_y(value[i+1])
            return self


    def __add__(self, value):
        if len(value) == 4:
            return Vec4(self[0]+value[0], self[1]+value[1], self[2]+value[2], self[3]+value[3])

        if len(value) == 3:
            return Vec4(self[0]+value[0], self[1]+value[1], self[2]+value[2], self[3])

        elif len(value) == 2:
            return Vec4(self[0]+value[0], self[1]+value[1], self[2], self[3])



    def __mul__(self, value):
        if isinstance(value, int | float | complex):
            return Vec4(*(e*value for e in self))

        return Vec4(self[0]*value[0], self[1]*value[1], self[2]*value[2], self[3]*value[3])


    __rmul__ = __mul__


    def __truediv__(self, value):
        if isinstance(value, int | float | complex):
            return Vec4(*(e/value for e in self))

        return Vec4(self[0]/value[0], self[1]/value[1], self[2]/value[2], self[3]/value[3])


    def __abs__(self):
        return Vec4(*[abs(e) for e in self])

if __name__ == '__main__':
    from ursina.ursinastuff import _test
    _test(Vec4(1,0,0,0) * 2 == Vec4(2,0,0,0))
    _test(Vec4(1,0,1,1) * Vec4(2,1,2,3) == Vec4(2,0,2,3))

    def _test_Vec4_plus_equal():
        b = Vec4(1.252352324, 0, 1, .2)
        b += Vec4(0,1)
        return b == Vec4(1.252352324, 1, 1, .2)
    _test(_test_Vec4_plus_equal)



from ursina.shader import Shader

normals_shader = Shader(name='normals_shader',language=Shader.GLSL,
vertex = '''
#version 140
uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelMatrix;
in vec4 p3d_Vertex;
in vec3 p3d_Normal;
out vec3 world_normal;

void main() {
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    world_normal = normalize(mat3(p3d_ModelMatrix) * p3d_Normal);
}
''',

fragment='''
#version 130

uniform vec4 p3d_ColorScale;
in vec2 texcoord;
out vec4 fragColor;
in vec3 world_normal;


void main() {
    fragColor = vec4(world_normal*0.5+0.5, 1);
}

''',
geometry='',
)



if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import AzureSphere, GrayPlane, WhiteCube
    app = Ursina()
    window.color=color.black

    # e = Entity(model='sphere', shader=normals_shader)
    # e.setShaderInput('transform_matrix', e.getNetTransform().getMat())
    shader = normals_shader

    a = WhiteCube(shader=shader)
    b = AzureSphere(rotation_y=180, x=3)
    b.shader = shader
    # AzureSphere(shader=a.shader, y=2)
    GrayPlane(scale=10, y=-2, texture='shore')

    Sky(color=color.light_gray)
    EditorCamera()

    def update():
        b.rotation_z += 1
        b.rotation_y += 1
        b.rotation_x += 1
        # a.rotation_x += 1
    # EditorCamera()

    app.run()



import sys
import warnings


# can removed when Python3.13 releases
class deprecated:
    def __init__(self, message: str, /, *, category: type[Warning] | None = DeprecationWarning, stacklevel: int = 1, ) -> None:
        if not isinstance(message, str):
            raise TypeError(f"Expected an object of type str for 'message', not {type(message).__name__!r}")
        self.message = message
        self.category = category
        self.stacklevel = stacklevel

    def __call__(self, arg, /):
        msg = self.message
        category = self.category
        stacklevel = self.stacklevel
        if category is None:
            arg.__deprecated__ = msg
            return arg
        elif isinstance(arg, type):
            import functools
            from types import MethodType

            original_new = arg.__new__

            @functools.wraps(original_new)
            def __new__(cls, *args, **kwargs):
                if cls is arg:
                    warnings.warn(msg, category=category, stacklevel=stacklevel + 1)
                if original_new is not object.__new__:
                    return original_new(cls, *args, **kwargs)
                # Mirrors a similar check in object.__new__.
                elif cls.__init__ is object.__init__ and (args or kwargs):
                    raise TypeError(f"{cls.__name__}() takes no arguments")
                else:
                    return original_new(cls)

            arg.__new__ = staticmethod(__new__)

            original_init_subclass = arg.__init_subclass__
            # We need slightly different behavior if __init_subclass__
            # is a bound method (likely if it was implemented in Python)
            if isinstance(original_init_subclass, MethodType):
                original_init_subclass = original_init_subclass.__func__

                @functools.wraps(original_init_subclass)
                def __init_subclass__(*args, **kwargs):
                    warnings.warn(msg, category=category, stacklevel=stacklevel + 1)
                    return original_init_subclass(*args, **kwargs)

                arg.__init_subclass__ = classmethod(__init_subclass__)
            # Or otherwise, which likely means it's a builtin such as
            # object's implementation of __init_subclass__.
            else:
                @functools.wraps(original_init_subclass)
                def __init_subclass__(*args, **kwargs):
                    warnings.warn(msg, category=category, stacklevel=stacklevel + 1)
                    return original_init_subclass(*args, **kwargs)

                arg.__init_subclass__ = __init_subclass__

            arg.__deprecated__ = __new__.__deprecated__ = msg
            __init_subclass__.__deprecated__ = msg
            return arg
        elif callable(arg):
            import functools

            @functools.wraps(arg)
            def wrapper(*args, **kwargs):
                warnings.warn(msg, category=category, stacklevel=stacklevel + 1)
                return arg(*args, **kwargs)

            arg.__deprecated__ = wrapper.__deprecated__ = msg
            return wrapper
        else:
            raise TypeError(
                "@deprecated decorator with non-None category must be applied to "
                f"a class or callable, not {arg!r}"
            )


_DEPRECATED_MSG = "{name!r} is deprecated and slated for removal in Python {remove}"

def _deprecated(name, message=_DEPRECATED_MSG, *, remove, _version=sys.version_info):
    remove_formatted = f"{remove[0]}.{remove[1]}"
    if (_version[:2] > remove) or (_version[:2] == remove and _version[3] != "alpha"):
        msg = f"{name!r} was slated for removal after Python {remove_formatted} alpha"
        raise RuntimeError(msg)
    else:
        msg = message.format(name=name, remove=remove_formatted)
        warnings.warn(msg, DeprecationWarning, stacklevel=3)



from ursina import *


app = Ursina()
class MenuButton(Button):
    def __init__(self, text='', **kwargs):
        super().__init__(text, scale=(.25, .075), highlight_color=color.azure, **kwargs)

        for key, value in kwargs.items():
            setattr(self, key ,value)


# button_size = (.25, .075)
button_spacing = .075 * 1.25
menu_parent = Entity(parent=camera.ui, y=.15)
main_menu = Entity(parent=menu_parent)
load_menu = Entity(parent=menu_parent)
options_menu = Entity(parent=menu_parent)

state_handler = Animator({
    'main_menu' : main_menu,
    'load_menu' : load_menu,
    'options_menu' : options_menu,
    }
)


# main menu content
main_menu.buttons = [
    MenuButton('start', on_click=Func(setattr, state_handler, 'state', 'load_menu')),
    MenuButton('options', on_click=Func(setattr, state_handler, 'state', 'options_menu')),
    MenuButton('quit', on_click=Sequence(Wait(.01), Func(application.quit))),
]
for i, e in enumerate(main_menu.buttons):
    e.parent = main_menu
    e.y = (-i-2) * button_spacing


def start_game():
    menu_parent.enabled = False
    import tic_tac_toe

# load menu content
for i in range(3):
    MenuButton(parent=load_menu, text=f'Empty Slot {i}', y=-i * button_spacing, on_click=start_game)

load_menu.back_button = MenuButton(parent=load_menu, text='back', y=((-i-2) * button_spacing), on_click=Func(setattr, state_handler, 'state', 'main_menu'))


# options menu content
review_text = Text(parent=options_menu, x=.275, y=.25, text='Preview text', origin=(-.5,0))
for t in [e for e in scene.entities if isinstance(e, Text)]:
    t.original_scale = t.scale

text_scale_slider = Slider(0, 2, default=1, step=.1, dynamic=True, text='Text Size:', parent=options_menu, x=-.25)
def set_text_scale():
    for t in [e for e in scene.entities if isinstance(e, Text) and hasattr(e, 'original_scale')]:
        t.scale = t.original_scale * text_scale_slider.value
text_scale_slider.on_value_changed = set_text_scale


def update_volume_for_currently_playing():
    for e in scene.entities:
        if isinstance(e, Audio):
            e.volume = e.volume

volume_slider = Slider(0, 1, default=Audio.volume_multiplier, step=.1, text='Master Volume:', parent=options_menu, x=-.25,
    on_value_changed = lambda: (
        setattr(Audio, 'volume_multiplier', volume_slider.value),
        update_volume_for_currently_playing()
    )
)
# @(lambda f:setattr(volume_slider, 'on_value_changed', f))
# def anon():
#     Audio.volume_multiplier = volume_slider.value
#     update_volume_for_currently_playing()

# volume_slider.on_value_changed = set_volume_multiplier

options_back = MenuButton(parent=options_menu, text='Back', x=-.25, origin_x=-.5, on_click=Func(setattr, state_handler, 'state', 'main_menu'))

for i, e in enumerate((text_scale_slider, volume_slider, options_back)):
    e.y = -i * button_spacing


# animate the buttons in nicely when changing menu
for menu in (main_menu, load_menu, options_menu):
    def animate_in_menu(menu=menu):
        for i, e in enumerate(menu.children):
            e.original_x = e.x
            e.x += .1
            e.animate_x(e.original_x, delay=i*.05, duration=.1, curve=curve.out_quad)

            e.alpha = 0
            e.animate('alpha', .7, delay=i*.05, duration=.1, curve=curve.out_quad)

            if hasattr(e, 'text_entity'):
                e.text_entity.alpha = 0
                e.text_entity.animate('alpha', 1, delay=i*.05, duration=.1)

    menu.on_enable = animate_in_menu


background = Entity(parent=menu_parent, model='quad', texture='shore', scale=(camera.aspect_ratio,1), color=color.dark_gray, z=1, world_y=0)
# application.asset_folder = application.asset_folder.parent
# Audio('chillstep_1.ogg')
app.run()



from ursina import color
from ursina.shader import Shader
from ursina.vec3 import Vec3

toon_shader = Shader(name='toon_shader', language=Shader.GLSL, vertex = '''#version 140
uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelMatrix;

in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 texcoord;

in vec3 p3d_Normal;
out vec3 world_normal;

void main() {
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    texcoord = p3d_MultiTexCoord0;
    world_normal = normalize(mat3(p3d_ModelMatrix) * p3d_Normal);
}
''',
fragment='''
#version 120

uniform vec3 lightDir;
varying vec3 world_normal;

void main() {
    float intensity;
    vec4 color;

    intensity = dot(lightDir, world_normal);
    if (intensity > 0.95) color = vec4(1.0, 0.5, 0.5, 1.0);
    else if (intensity > 0.5) color = vec4(0.6, 0.3, 0.3, 1.0);
    else if (intensity > 0.25) color = vec4(0.4, 0.2, 0.2, 1.0);
    else color = vec4(0.2, 0.1, 0.1, 1.0);

    gl_FragColor = color;
}
''',
default_input={
    # 'texture_scale' : Vec2(1,1),
    # 'texture_offset' : Vec2(0.0, 0.0),
    'lightDir' : Vec3(1,-1,1)
}
)


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    shader = toon_shader
    e = Entity(model='sphere', y=2, color=color.azure, shader=shader)
    e.model.generate_normals(smooth=True)
    Entity(model='plane', scale=8, shader=shader)
    EditorCamera()
    def update():
        if held_keys["d"]: e.rotation_y += 4
        if held_keys["a"]: e.rotation_y -= 4
    app.run()



from ursina import *


class Plane(Mesh):
    def __init__(self, subdivisions=(1,1), mode='triangle', **kwargs):

        self.vertices, self.triangles = list(), list()
        self.uvs = list()

        w, h = subdivisions
        i = 0

        for z in range(h+1):
            for x in range(w+1):
                self.vertices.append(Vec3((x/w)-.5, 0, (z/h)-.5))
                self.uvs.append((x/w, z/h))

                if x > 0 and z > 0:
                    self.triangles.append((i, i-1, i-w-2, i-w-1))

                i += 1

        super().__init__(vertices=self.vertices, triangles=self.triangles, uvs=self.uvs, mode=mode, **kwargs)



if __name__ == '__main__':
    app = Ursina()

    front =  Entity(model=Plane(subdivisions=(3,6)), texture='brick', rotation_x=-90)
    wireframe_renderer =  duplicate(front, wireframe=True, color=color.azure, always_on_top=True)

    _ed = EditorCamera()
    Entity(model='cube', color=color.green, scale=.05)
    app.run()



from ursina.editor.level_editor import *


class PipeEditor(Entity):
    def __init__(self, points=[Vec3(0,0,0), Vec3(0,1,0)], **kwargs):
        super().__init__(original_parent=LEVEL_EDITOR, selectable=True, name='Pipe', **kwargs)
        LEVEL_EDITOR.entities.append(self)
        self._point_gizmos = LoopingList([Entity(parent=self, original_parent=self, position=e, selectable=False, name='PipeEditor_point', is_gizmo=True) for e in points])
        self.model = Pipe()
        self.edit_mode = False
        self.add_collider = False
        self.generate()


    def generate(self):
        self.model = Pipe(
            path = [e.get_position(relative_to=self) for e in self._point_gizmos],
            thicknesses = [e.scale.xz for e in self._point_gizmos]
        )
        self.texture = 'grass'

        if self.add_collider:
            self.collider = self.model


    def __deepcopy__(self, memo):
        return eval(repr(self))


    @property
    def points(self):
        return [e.position for e in self._point_gizmos]

    @property
    def edit_mode(self):
        return self._edit_mode

    @edit_mode.setter
    def edit_mode(self, value):
        # print('set edit mode', value)
        self._edit_mode = value
        if value:
            [setattr(e, 'selectable', False) for e in LEVEL_EDITOR.entities if not e == self]
            for e in self._point_gizmos:
                if not e in LEVEL_EDITOR.entities:
                    LEVEL_EDITOR.entities.append(e)

            [setattr(e, 'selectable', True) for e in self._point_gizmos]
        else:
            [LEVEL_EDITOR.entities.remove(e) for e in self._point_gizmos if e in LEVEL_EDITOR.entities]
            [setattr(e, 'selectable', True) for e in LEVEL_EDITOR.entities]
            if True in [e in LEVEL_EDITOR.selection for e in self._point_gizmos]: # if point is selected when exiting edit mode, select the poke shape
                LEVEL_EDITOR.selection = [self, ]

        LEVEL_EDITOR.render_selection()


    def input(self, key):
        combined_key = input_handler.get_combined_key(key)
        if combined_key == 'tab':
            if self in LEVEL_EDITOR.selection or True in [e in LEVEL_EDITOR.selection for e in self._point_gizmos]:
                self.edit_mode = not self.edit_mode
                return

        elif key == '+' and len(LEVEL_EDITOR.selection) == 1 and LEVEL_EDITOR.selection[0] in self._point_gizmos:
            print('add point')
            i = self._point_gizmos.index(LEVEL_EDITOR.selection[0])
        
            new_point = Entity(parent=self, original_parent=self, position=lerp(self._point_gizmos[i].position, self._point_gizmos[i+1].position, .5), selectable=True, is_gizmo=True)
            LEVEL_EDITOR.entities.append(new_point)
            self._point_gizmos.insert(i+1, new_point)
            LEVEL_EDITOR.render_selection()

        elif key == 'space':
            self.generate()

        elif self.edit_mode and key.endswith(' up'):
            invoke(self.generate, delay=3/60)


if __name__ == '__main__':

    app = Ursina(borderless=False)

    level_editor = LevelEditor()
    level_editor.goto_scene(0,0)
    # cube = WhiteCube(selectable=True)
    level_editor.entities.append(PipeEditor())
    app.run()



from ursina import *

app = Ursina()

rotation_resetter = Entity()
cube = Entity(parent=rotation_resetter, model='cube', texture='white_cube')


def update():
    rotation_resetter.rotation_x += 100 * (held_keys['a'] - held_keys['d']) * time.dt
    rotation_resetter.rotation_z += 100 * (held_keys['w'] - held_keys['s']) * time.dt

    cube.rotation = cube.world_rotation
    rotation_resetter.rotation = (0,0,0)

EditorCamera()

app.run()



from ursina import Button, Entity, Func, Mesh, Plane, ThinSlider, Vec2, Vec3, camera, color, copy, generate_properties_for_class, make_gradient
from ursina.prefabs.color_picker import ColorPicker

gradient_editor_arrow = Mesh(vertices=[(v+Vec3(0,-.4,0))*Vec3(.0175,.025,1) for v in (Vec3(0,0,0),Vec3(-.5,.5,0),Vec3(.5,.5,0),Vec3(-.5,.75,0),Vec3(.5,.75,0))], triangles=(0,2,1,1,2,4,4,3,1))

@generate_properties_for_class()
class GradientEditor(Entity):
    default_values = dict(parent=camera.ui, )

    def __init__(self, value={'0':color.hex('#ffffffff'), '32':color.hex('#ffffffff'),}, resolution=32, on_value_changed:callable=None, **kwargs):
        super().__init__(**(__class__.default_values | kwargs))

        self.color_picker = ColorPicker(parent=self, scale=.75, enabled=False, show_exit_button=True)
        self.color_picker.exit_button.on_click = self.stop_editing_color

        self.resolution = resolution
        self.sliders = []
        for i in range(3):
            slider = ThinSlider(parent=self, x=-.25, step=1, min=0, max=resolution-1, default=lerp(0,resolution,i/(3-1)), dynamic=True)
            self.sliders.append(slider)
            slider.knob.color=color.random_color()
            def _set_value():
                self.preview_gradient()
                if self.on_value_changed:
                    self.on_value_changed()

            slider.on_value_changed = _set_value
            slider.bg.enabled = False
            slider.knob.text_entity.enabled = False
            slider.knob.model = copy(gradient_editor_arrow)
            slider.knob.highlight_color = slider.knob.color
            slider.knob.on_double_click = Func(self.start_editing_color, slider)

        self.gradient_renderer = Entity(parent=self, model=Plane((resolution,1)), rotation_x=-90, scale=Vec3(.5,1,.05), y=-.035)
        self.copy_button = Button(parent=self, scale=Vec2(.05,.025)*1, text='copy', text_size=.5, origin=(.5,-.5), x=-.25-.01, y=-.06, on_click=self.copy)

        self.on_value_changed = on_value_changed    # set this to a function you want to be called when the slider changes
        self.value = value


    def start_editing_color(self, slider):
        if self.color_picker.enabled:
            self.stop_editing_color()

        self.color_picker.enabled = True
        self.color_picker.position = slider.knob.position + Vec2(-.25,.25*.75)
        self.color_picker.value = slider.knob.color

        def _set_color():
            slider.knob.color = self.color_picker.color
            self.value = self.value
        self.color_picker.on_value_changed = _set_color


    def stop_editing_color(self):
        self.color_picker.enabled = False
        self.color_picker.on_value_changed = None


    def preview_gradient(self):
        self.gradient_renderer.model.colors = [self.gradient[int((v.x+.5)*self.resolution)] for v in self.gradient_renderer.model.vertices]
        self.gradient_renderer.model.generate()

    def gradient_getter(self):  # get the final calculated gradient as a list of 32 color
        index_color_dict = self.value
        index_color_dict['0'] = min(self.sliders, key=lambda s:s.value).knob.color
        index_color_dict[str(self.resolution)] = max(self.sliders, key=lambda s:s.value).knob.color
        return make_gradient(index_color_dict)


    def value_getter(self):
        return {str(slider.value) : slider.knob.color for slider in self.sliders}

    def value_setter(self, value, call_on_value_changed=True):
        for i, (key, col) in enumerate(value.items()):
            if isinstance(col, str):
                col = color.hex(col)

            self.sliders[i].knob.color = col
            self.sliders[i].value_setter(int(key), call_on_value_changed=False)

        self._value = value
        self.preview_gradient()
        if self.on_value_changed and call_on_value_changed:
            self.on_value_changed()

    def copy(self):
        # print('copied:', {key:color.rgb_to_hex(*col) for key, col in self.value.items()})
        import pyperclip
        pyperclip.copy(str({key:color.rgb_to_hex(*col) for key, col in self.value.items()}))


if __name__ == '__main__':
    from ursina import Ursina
    app = Ursina()
    GradientEditor()
    app.run()


from ursina import *


class PauseMenu(Entity):
    def __init__(self, **kwargs):
        super().__init__(ignore_paused=True, **kwargs)
        self.menu = Entity(parent=camera.ui, enabled=False)
        self.bg = Entity(parent=self.menu, model='quad', color=color.black, alpha=.5, scale=3)
        self.pause_text = Text(parent=self.menu, text='PAUSED', origin=(0,0), scale=2)
        self.lock_mouse_on_resume = False

    def on_destroy(self):
        destroy(self.menu)

    def input(self, key):
        if key == 'escape':
            if not application.paused:
                self.lock_mouse_on_resume = mouse.locked
                mouse.locked = False
            else:
                mouse.locked = self.lock_mouse_on_resume

            application.paused = not application.paused # Pause/unpause the game.
            self.menu.enabled = application.paused     # Also toggle "PAUSED" graphic.



if __name__ == '__main__':
    app = Ursina()

    PauseMenu()

    app.run()



from ursina import *



app = Ursina(forced_aspect_ratio=.6)


bg = Entity(model='quad', scale=(30, 50), texture='grass', color=hsv(0,0,.2))
player = Entity(model=Circle(3), color=color.azure, speed=8, y=-.4, z=-1)
player.bullet_renderer = Entity(model=Mesh(mode='point', thickness=.2), texture='circle', color=color.yellow)

scene.fog_density = (10,50)
ec = EditorCamera(rotation_x=-20)


def shoot():
    player.bullet_renderer.model.vertices.append(player.position)

shoot_cooldown = .1
shoot_sequence = Sequence(Func(shoot), Wait(shoot_cooldown), loop=True)


def update():
    move_direction = Vec2(held_keys['d']-held_keys['a'], held_keys['w']-held_keys['s']).normalized()
    player.position += move_direction * player.speed * time.dt
    bg.texture_offset += Vec2(0, time.dt)

    for i, bullet in enumerate(player.bullet_renderer.model.vertices):
        player.bullet_renderer.model.vertices[i] += Vec3(0, time.dt * 10, 0)
        for enemy in enemies:
            if distance_2d(bullet, enemy) < .5:
                enemy.hp -= 1
                enemy.blink(color.white)
                if enemy.hp <= 0:
                    enemies.remove(enemy)
                    destroy(enemy)
                    # todo: add explosion particles and sound effect
                player.bullet_renderer.model.vertices.remove(bullet)

                print('a')

    if len(player.bullet_renderer.model.vertices):
        player.bullet_renderer.model.vertices = player.bullet_renderer.model.vertices[-100:]  # max bullets

    player.bullet_renderer.model.generate()


def input(key):
    if key == 'space':
        shoot_sequence.start()
    if key == 'space up':
        shoot_sequence.paused = True




enemies = []
enemy = Entity(model=Circle(3), rotation_z=180, position=(0, 16), color=color.red, z=-1, speed=3, hp=5)
enemies.append(enemy)

def enemy_update():
    for e in enemies:
        e.position += e.up * enemy.speed * time.dt



enemy_handler = Entity(update=enemy_update)

app.run()



from panda3d.core import Texture as PandaTexture
from panda3d.core import TexturePool
from panda3d.core import SamplerState
from panda3d.core import Filename
from panda3d.core import PNMImage
from pathlib import Path
from ursina.vec2 import Vec2
from ursina import color
from ursina.ursinamath import clamp
from ursina.array_tools import Array2D, enumerate_2d
from ursina.scripts.property_generator import generate_properties_for_class

@generate_properties_for_class()
class Texture():

    default_filtering = None      # options: None / 'bilinear' / 'mipmap'

    def __init__(self, value, filtering='default'):

        if isinstance(value, str) and value.startswith('data:image/png;base64'):
            import base64
            from PIL import Image
            from io import BytesIO
            base64_data = value.lstrip('data:image/png;base64')   # remove prefix
            value = Image.open(BytesIO(base64.b64decode(base64_data))).convert('RGBA')

        elif isinstance(value, str):
            value = Path(value)

        if isinstance(value, Path):
            self.path = Path(value)
            self._texture = TexturePool.loadTexture(Filename.fromOsSpecific(str(value)))
            self._cached_image = None   # for get_pixel() method

        elif isinstance(value, PandaTexture):
            self._texture = value

        else:
            from PIL import Image
            image = value.convert('RGBA')
            self._texture = PandaTexture()
            self._texture.setup2dTexture(image.width, image.height, PandaTexture.TUnsignedByte, PandaTexture.FRgba)
            self._texture.setRamImageAs(image.transpose(Image.FLIP_TOP_BOTTOM).tobytes(), image.mode)
            self._cached_image = image   # for get_pixel() method
            # self._cached_image = image.transpose(Image.FLIP_TOP_BOTTOM)
            self.path = None

        if filtering == 'default':
            filtering = Texture.default_filtering      # None/'bilinear'/'mipmap' default: 'None'
        self.filtering = filtering


    @staticmethod
    def new(size, color=(255,255,255)):
        img = PNMImage(*size)
        if len(color) == 4:
            img.addAlpha()

        img.fill(*color)
        panda_tex = PandaTexture('texture')
        panda_tex.load(img)
        return Texture(panda_tex)


    def name_getter(self):
        try:
            return self.path.name
        except:
            return f'PIL_texture_{self.size}'


    def __str__(self):
        return self.name


    def size_getter(self):
        return Vec2(self.width, self.height)

    def width_getter(self):
        if self._cached_image:
            return self._cached_image.size[0]
        elif self._texture.getOrigFileXSize() > 0:
            return self._texture.getOrigFileXSize()
        return 0

    def height_getter(self):
        if self._cached_image:
            return self._cached_image.size[1]
        elif self._texture.getOrigFileYSize() > 0:
            return self._texture.getOrigFileYSize()
        return 0


    def pixels_getter(self):
        pixels = Array2D(*self.size, default_value=None)
        for (x,y), value in enumerate_2d(pixels):
            pixels[x][y] = self.get_pixel(x, y)

        return pixels

    def filtering_setter(self, value):
        # print('setting filtering:', value)
        if value in (None, False, 'nearest', 'nearest neighbor', 'point'):
            self._texture.setMagfilter(SamplerState.FT_nearest)
            self._texture.setMinfilter(SamplerState.FT_nearest)
            self._filtering = False
        elif value in (True, 'linear', 'bilinear'):
            self._texture.setMagfilter(SamplerState.FT_linear)
            self._texture.setMinfilter(SamplerState.FT_linear)
            self._filtering = True
        elif value == 'mipmap':
            self._texture.setMinfilter(SamplerState.FT_linear_mipmap_linear)
            self._filtering = 'mipmap'


    def repeat_setter(self, value):
        self._repeat = value
        self._texture.setWrapU(value)
        self._texture.setWrapV(value)


    def get_pixel(self, x, y):
        try:
            if not self._cached_image:
                from PIL import Image
                self._cached_image = Image.open(self.path)


            col = self._cached_image.getpixel((x, self.height-y-1))
            if self._cached_image.mode == 'LA':
                col = (col[0], col[0], col[0], col[1])

            if self._cached_image.mode == 'L':
                col = (col[0], col[0], col[0])

            return color.rgba32(*col)
        except Exception as e:
            print(e)
            return None


    def get_pixels(self, start, end):
        start = (clamp(start[0], 0, self.width), clamp(start[1], 0, self.width))
        end = (clamp(end[0], 0, self.width), clamp(end[1], 0, self.width))
        pixels = []

        for y in range(start[1], end[1]):
            for x in range(start[0], end[0]):
                pixels.append(self.get_pixel(x,y))

        return pixels

    def set_pixel(self, x, y, color):
        if not self._cached_image:
            from PIL import Image
            self._cached_image = Image.open(self.path)

        self._cached_image.putpixel((x, self.height-y-1), tuple(int(e*255) for e in color))


    def apply(self):
        from PIL import Image
        if not self._cached_image:
            self._cached_image = Image.open(self.path)

        self._texture.setRamImageAs(self._cached_image.transpose(Image.FLIP_TOP_BOTTOM).tobytes(), self._cached_image.mode)
        # self._texture.setRamImageAs(self._cached_image.tobytes(), self._cached_image.mode)


    def save(self, path):
        if not self._cached_image:
            from PIL import Image
            self._cached_image = Image.open(self.path)

        self._cached_image.save(path)


    def to_PIL_Image(self):
        from PIL import Image
        # Ensure the texture data is available
        # self._texture.prepareRamImage()

        # Get the size and format
        width = self._texture.getXSize()
        height = self._texture.getYSize()

        # Get the raw image data as bytes in RGB format
        data = self._texture.getRamImageAs("RGB")
        if data is None:
            raise RuntimeError("Texture has no RAM image data")

        # Convert from bottom-up to top-down image
        image = Image.frombytes("RGB", (width, height), data.getData())
        image = image.transpose(Image.FLIP_TOP_BOTTOM)
        return image


    def __repr__(self):
        return self.name


if __name__ == '__main__':
    from ursina import *
    from ursina import texture_importer
    app = Ursina()
    '''
        The Texture class rarely used manually but usually instantiated
        when assigning a texture to an Entity
        texture = Texture(path / PIL.Image / panda3d.core.Texture)

        A texture file can be a .png, .jpg or .psd.
        If it's a .psd it and no compressed version exists, it will compress it automatically.
    '''
    e = Entity(model='quad', texture='test_tileset')
    e.texture.set_pixel(0, 2, color.blue)
    e.texture.apply()
    #
    # for y in range(e.texture.height):
    #     for x in range(e.texture.width):
    #         if e.texture.get_pixel(x,y) == color.blue:
    #             print('found blue pixel at:', x, y)

    # test
    application.asset_folder = Path(r'C:\sync\high resolution images')
    e = Entity(model='quad')
    # from PIL import Image

    # from ursina.prefabs.memory_counter import MemoryCounter
    # MemoryCounter()
    def input(key):
        if key == 'a':
            # img = Image.open(r'C:\sync\high resolution images\tesla_city.png')
            # print('-------', img)
            # e.texture = Texture(img)
            e.texture = 'tesla_city'

        if key == 'space':
            # if not 'tesla_city' in texture_importer.imported_textures:
            #     return
            # print('del texture')
            # del texture_importer.imported_textures['tesla_city']
            # del e.texture
            # tex = e.texture._texture
            t = e.texture._texture
            # e.texture._texture = None
            e.texture = None
            t.releaseAll()
            # print(t.releaseAll())
            t.clearRamImage()
            # e.texture._texture.clear()
            # e.texture._texture = None
            # e.texture = None
            # destroy(e)
        if key == 'p':
            for key, value in texture_importer.imported_textures.items():
                print(key, value)

    e.texture = 'test_tileset'
    # e.texture.set_pixels([e for e in e.texture.get_pixels()])
    e.texture.apply()

    # tex = Texture.new((512,512), (255,0,0))

    pixels = e.texture.pixels
    new_grid = Array2D(width=pixels.width, height=pixels.height)
    print('w:', pixels.width, 'h:', pixels.height)
    for (x,y), value in enumerate_2d(pixels):
        new_grid[x][y] = int(color.rgba32(*value).v > .5)


    texture_from_base64_string = Entity(model='cube', y=1.5, scale=1, texture=Texture('data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAQAAAAECAIAAAAmkwkpAAAAJ0lEQVR4nGK5d/gdAwODArcEkGRiQAKMmut0gdTX/YroMoAAAAD//8caBbV8Qu6pAAAAAElFTkSuQmCC'))
    EditorCamera()

    # print(new_grid)
    app.run()



from pathlib import Path
from copy import copy
import builtins
import importlib
from ursina import application
from ursina.texture import Texture


imported_textures = dict()
file_types = ('.tif', '.jpg', '.jpeg', '.png', '.gif')
textureless = False

use_thumbhash = False
thumbhashes = None
thumbhash_path = application.textures_compressed_folder / 'thumbhashes.json'


def load_texture(name, folder:Path=None, use_cache=True, filtering='default'):
    if textureless and '*' not in name:
        return Texture(application.internal_textures_folder/'white_cube.png')

    if use_cache and name in imported_textures:
        return copy(imported_textures[name])

    if use_thumbhash:
        global thumbhashes
        import thumbhash
        import json
        if thumbhashes is None:
            with thumbhash_path.open('r') as file:
                thumbhashes = json.load(file)

        if name in thumbhashes:
            image = thumbhash.thumbhash_to_image(thumbhashes[name])
            tex = Texture(image, filtering='bilinear')
            imported_textures[name] = tex
            return tex

    if folder is not None:
        if not isinstance(folder, Path):
            raise TypeError(f'folder must be a Path, not a {type(folder)}')
        _folders = (folder,)

    else:
        _folders = (application.textures_compressed_folder, application.asset_folder, application.internal_textures_folder)


    if name.endswith('.mp4'):
        for folder in _folders:
            for filename in folder.glob('**/' + name):
                # print('loaded movie texture:', filename)
                return builtins.loader.loadTexture(filename.resolve())


    for folder in _folders:
        if '.' in name: # got name with file extension
            for filename in folder.glob('**/' + name):
                t = Texture(filename.resolve(), filtering=filtering)
                imported_textures[name] = t
                return t

        for filename in folder.glob('**/' + name + '.*'): # no file extension given, so try all supported
            if filename.suffix in file_types:
                # print('found:', filename)
                t = Texture(filename.resolve(), filtering=filtering)
                imported_textures[name] = t
                return t

    if application.development_mode and importlib.util.find_spec('psd_tools'):
        from psd_tools import PSDImage

        for folder in _folders:
            for filename in folder.glob('**/' + name + '.psd'):
                print('found uncompressed psd, compressing it...')
                compress_textures(name)
                return load_texture(name)

    imported_textures[name] = None  # prevent searching for the same missing texture multiple times
    return None



def compress_textures(name=''):
    try:
        from PIL import Image
    except Exception as e:
        return e


    if not application.textures_compressed_folder.exists():
        application.textures_compressed_folder.mkdir()


    file_type = '.*'
    if '.' in name:
        file_type = ''

    # print('searching for texture:', name + file_type)
    for f in application.asset_folder.glob('**/' + name + file_type):

        if '\\textures_compressed\\' in str(f) or f.suffix not in ('.psd', '.png', '.jpg', '.jpeg', '.gif'):
            continue
        # print('  found:', f)
        image = None
        if f.suffix == '.psd':
            try:
                from psd_tools import PSDImage
            except (ModuleNotFoundError, ImportError):
                print('info: psd-tools3 not installed')
                return None

            image = PSDImage.load(f)
            image = image.as_PIL()
        # elif f.suffix == '.png':
        #     image = Image.open(f)

        if not image:
            return False
        # print(max(image.size))
        # print('............', image.mode)
        if image and image.mode != 'RGBA' and max(image.size) > 512:
            image.save(
                application.textures_compressed_folder / (Path(f).stem + '.jpg'),
                'JPEG',
                quality=80,
                optimize=True,
                progressive=True
                )
            print('    compressing to jpg:', application.textures_compressed_folder / (f.stem + '.jpg'))
            continue
        else:
            image.save(application.textures_compressed_folder / (f.stem + '.png'), 'PNG')
            print('    compressing to png:', application.textures_compressed_folder / (f.stem + '.png'))


def generate_thumbhashes():
    import json
    thumbhashes = dict()
    # for suffix in ('jpg', 'jpeg', 'png', 'tif'):
    for image_path in application.asset_folder.glob('**/*.*'):
        if image_path.suffix not in ('.jpg', '.jpeg', '.png', '.tif'):
            continue
        if image_path.stem in thumbhashes:
            continue

        print('making thumbhash for:', image_path.name)
        with image_path.open('rb') as image_file:
            thumbhashes[image_path.stem] = image_to_thumbhash(image_file)
            # image = thumbhash.thumbhash_to_image(thumbhashes[name], base_size=128)
            # tex = Texture(image)

    if not application.textures_compressed_folder.exists():
        application.textures_compressed_folder.mkdir()

    with thumbhash_path.open('w') as file:
        json.dump(thumbhashes, file, indent=0)


if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    Entity(model='quad', texture='white_cube')
    app.run()



from ursina import camera, color, window
from ursina.shader import Shader
from ursina.ursinastuff import Func
from ursina.vec2 import Vec2
from ursina.vec3 import Vec3

unlit_points_shader = Shader(name='unlit_points_shader', language=Shader.GLSL,
vertex='''
#version 330 core

layout(location = 0) in vec4 p3d_Vertex;
layout(location = 1) in vec4 color;  // Vertex color input
layout(location = 2) in vec2 p3d_MultiTexCoord0;

uniform mat4 p3d_ModelViewMatrix;
uniform mat4 p3d_ProjectionMatrix;
uniform mat4 p3d_ModelMatrix; // Model matrix for converting to world space

out vec4 vertex_color;
out vec2 texCoord;
out vec3 vertex_world_position;
out float point_size; // Output point size based on distance from the camera

void main() {
    // Calculate the world position of the vertex
    vertex_world_position = (p3d_ModelMatrix * p3d_Vertex).xyz;

    // Calculate the camera position by inverting the ModelView matrix and applying it to the origin
    vec4 camera_position = inverse(p3d_ModelViewMatrix) * vec4(0.0, 0.0, 0.0, 1.0);

    // Calculate the distance from the camera to the vertex
    float distance = length(camera_position.xyz - vertex_world_position);

    // Set the point size based on the distance (you can adjust this formula as needed)
    point_size = 1.0 / distance; // Example: Inverse distance scaling (adjust as needed)

    // Apply the model view matrix to position the vertex
    gl_Position = p3d_ModelViewMatrix * p3d_Vertex;

    // Pass color and texture coordinates to the next shader stage
    vertex_color = color;
    texCoord = p3d_MultiTexCoord0;
}

''',

geometry='''
#version 330 core
layout(points) in;
layout(triangle_strip, max_vertices = 4) out;

uniform mat4 p3d_ProjectionMatrix;
uniform mat4 p3d_ViewMatrix;
uniform float point_size;
uniform bool perspective_scaling;
uniform vec2 window_size;

out vec2 texCoord;
in vec4 vertex_color[];
out vec4 vert_col;
in vec3 vertex_world_position;



void main() {
    vec4 view_pos = gl_in[0].gl_Position;
    vec4 clip_pos = p3d_ProjectionMatrix * view_pos;


    float ndc_size = point_size / max(window_size.x, window_size.y) * 2.0;
    if (perspective_scaling) {
    }

    vec3 camera_right = vec3(1.,0,0);
    vec3 camera_up = vec3(0,1,0);
    vec3 right_offset = camera_right * (ndc_size * 0.5 * clip_pos.w);
    vec3 up_offset = camera_up * (ndc_size * 0.5 * clip_pos.w);

    vert_col = vertex_color[0];
    // Emit four vertices to form a quad facing the camera
    texCoord = vec2(0, 0);
    gl_Position = p3d_ProjectionMatrix * (view_pos - vec4(right_offset + up_offset, 0.0));
    EmitVertex();

    texCoord = vec2(1, 0);
    gl_Position = p3d_ProjectionMatrix * (view_pos + vec4(right_offset - up_offset, 0.0));
    EmitVertex();

    texCoord = vec2(0, 1);
    gl_Position = p3d_ProjectionMatrix * (view_pos - vec4(right_offset - up_offset, 0.0));
    EmitVertex();

    texCoord = vec2(1, 1);
    gl_Position = p3d_ProjectionMatrix * (view_pos + vec4(right_offset + up_offset, 0.0));
    EmitVertex();

    EndPrimitive();
}
''',


fragment='''
#version 330 core

in vec4 vert_col;
in vec2 texCoord;
uniform sampler2D p3d_Texture0;
out vec4 fragColor;

void main() {
    vec4 texColor = texture(p3d_Texture0, texCoord);
    fragColor = texColor * vert_col;
}

''',
default_input={
    'texture_scale' : Vec2(1,1),
    'texture_offset' : Vec2(0.0, 0.0),
    'point_size' : 1.0,

    'perspective_scaling': True,
},
continuous_input = {
    'window_size': Func(getattr, window, 'size'),
    }
)


if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    app = Ursina(vsync=1)
    camera.clip_plane_near= 1
    # window.color=color.black
    Entity(model='plane', scale=10, texture='grass')

    vertices=[Vec3(*(random.random()*10 for _ in range(3))) for i in range(100)]
    e = Entity(model=Mesh(mode=MeshModes.point, vertices=vertices, colors=[color.random_color() for _ in vertices], render_points_in_3d=True, thickness=50), shader=unlit_points_shader
    , texture='circle'
    )

    slider = ThinSlider(text='point_size', min=1, max=32, default=4)
    def _set_point_size():
        e.set_shader_input('point_size', slider.value)
        print('set point_size to:', slider.value, e.get_shader_input('point_size'))
    slider.on_value_changed = _set_point_size

    EditorCamera()
    app.run()



from copy import copy

from ursina import *


class _ConversationNode:
    __slots__ = ['index', 'indent_level', 'content', 'code', 'children', 'is_answer']

    def __str__(self):
        return f'{__class__.__name__}:\n    ' + '\n    '.join([f'{e} = {getattr(self, e)}' for e in __class__.__slots__])


class Conversation(Entity):

    def __init__(self, variables_object=None, **kwargs):
        super().__init__(parent=camera.ui, y=-.1)

        self.question = Button(parent=self, text_origin=(-.5,0), scale=(1,.1), model=Quad(radius=.5,aspect=1/.1), text='Question')
        self.question.text_entity.line_height = 1.25
        self.question.text_entity.position = (-.45, -.05)
        self.question.highlight_color = self.question.color
        self.more_indicator = Entity(parent=self.question, model=Circle(3), position=(.45,-.4,-1), rotation_z=180, color=color.azure, world_scale=.5, z=-1, enabled=False)
        self.more_indicator.animate_color(color.white, duration=1, loop=True, curve=curve.linear_boomerang)
        self.spacing = 4 * .02
        self.wordwrap = 65
        self.button_model = Quad(radius=.5, aspect=1/.075)
        self.variables_object = variables_object

        for key, value in kwargs.items():
            setattr(self, key ,value)

        self.answer_0 = Button(parent=self, text='answer_0', y=self.question.y-self.spacing-.025, scale=(1,.075), text_origin=(-.5,0), model=copy(self.button_model))
        self.answer_1 = Button(parent=self, text='answer_1', y=self.answer_0.y-self.spacing, scale=(1,.075), text_origin=(-.5,0), model=copy(self.button_model))
        self.answer_2 = Button(parent=self, text='answer_2', y=self.answer_1.y-self.spacing, scale=(1,.075), text_origin=(-.5,0), model=copy(self.button_model))

        self.buttons = (self.answer_0, self.answer_1, self.answer_2)
        for b in self.buttons:
            b.text_entity.line_height = 1.15
            b.text_entity.position = (-.45, 0)

        self.question_appear_sequence = None
        self.button_appear_sequence = None
        self.started = False



    def ask(self, node, question_part=0):
        # print(node)
        self.current_node = node
        self.question_part = question_part
        self.question.text = node.content[question_part]
        self.question.text_entity.wordwrap = self.wordwrap
        self.more_indicator.enabled = False
        self.question_appear_sequence = self.question.text_entity.appear()

        for b in self.buttons:
            b.enabled = False

        answers = []
        for i, child in enumerate(node.children):
            if self.variables_object and child.code and child.code.startswith('if'):
                try:
                    if not getattr(self.variables_object, child.code[3:]):
                        continue
                except Exception as e:
                    print('failed parsing conversation if statement:', e)

            answers.append(child)

        # multi page question
        if len(node.content) > 1 and self.question_part < len(node.content)-1:
            if self.question_part < len(node.content): # question not finished, so don't show answer buttons
                # print('question not finished')
                self.question_appear_sequence.append(Func(setattr, self.more_indicator, 'enabled', True))
                return

        self.button_appear_sequence = Sequence()
        invoke(self.button_appear_sequence.start, delay=self.question_appear_sequence.duration)

        if not node.children:
            self.buttons[0].text = '*leave*'
            self.buttons[0].on_click = Func(setattr, self, 'enabled', False)
            self.button_appear_sequence.append(Func(setattr, self.buttons[0], 'enabled', True))

        for i, child in enumerate(answers):
            self.button_appear_sequence.append(Wait(i*.15))
            self.button_appear_sequence.append(Func(setattr, self.buttons[i], 'enabled', True))
            self.buttons[i].text = child.content[0]
            self.buttons[i].text_entity.wordwrap = self.wordwrap

            def on_click(node=child):
                if not node.children:
                    print('end conversation')
                    self.enabled = False
                    return

                if node.code and not node.code.startswith('if '):
                    try:
                        if '+=' in node.code or '-=' in node.code or '*=' in node.code or '/=' in node.code:
                            var, operator, value = node.code.split()
                            original_value = getattr(self.variables_object, var)
                            data_type = type(original_value)
                            value = data_type(value)
                            if operator == '+=':    new_value = original_value + value
                            if operator == '-=':    new_value = original_value - value
                            if operator == '*=':    new_value = original_value * value
                            if operator == '/=':    new_value = original_value / value

                            setattr(self.variables_object, var, new_value)
                        print('executed code:', node.code)
                    except Exception as e:
                        print('failed executing code on node:', node, 'code:', node.code, 'error:', e)

                invoke(self.ask, node.children[0], 0, delay=.1)
                if len(node.children) > 1:
                    print('error at node:', n, '. node has multiple children, but should only have one (a question)')

            self.buttons[i].on_click = on_click



    def input(self, key):
        if key == 'left mouse down' or key == 'space' and not mouse.hovered_entity in self.buttons:
            self.next()


    def next(self):
        if not self.started:
            return

        if not self.question_appear_sequence.finished and self.question_appear_sequence.t > .1:
            self.question_appear_sequence.finish()
            if self.button_appear_sequence:
                self.button_appear_sequence.start()
            return

        if self.question_part < len(self.current_node.content)-1:
            self.ask(self.current_node, self.question_part+1)


    def start_conversation(self, conversation):
        self.conversation_nodes = self.parse_conversation(conversation)
        self.ask(self.conversation_nodes[0])
        self.started = True



    def parse_conversation(self, convo):
        convo = convo.strip()
        nodes = list()
        prev_node = None
        node_index = 0

        for i, l in enumerate(convo.split('\n')):
            if not l:
                continue

            indent_level = len(l) - len(l.lstrip())
            indent_level //= 4
            content, code = l.strip(), None
            if '(' in l:
                content, code = l.split('(')
                content = content.strip()
                code = code[:-1]
                is_answer = content.startswith('* ')
                # print('code:', codeK)

            if prev_node and prev_node.indent_level == indent_level:
                prev_node.content.append(content)
                prev_node = n
                continue

            n = _ConversationNode()
            n.index = node_index
            n.indent_level = indent_level
            n.is_answer = content.startswith('* ')
            if n.is_answer:
                content = content[2:]
            n.content = [content, ]
            n.children = list()
            n.code = code
            nodes.append(n)
            prev_node = n
            node_index += 1

            # look backwards through nodes to find current node's parent
            for j in range(node_index-1, -1, -1):
                if nodes[j].indent_level == n.indent_level-1:
                    nodes[j].children.append(n)
                    break


        return nodes



if __name__ == '__main__':
    app = Ursina()

    variables = Empty(
        evil=0,
        chaos=0,
        bar_mission_solved=False,
    )
    conversation = Conversation(variables_object=variables)
    # conversation.question.model = 'quad'
    # for b in conversation.buttons:
    #     b.model = 'quad'

    convo = dedent('''
    I'm looking for my sister. Can you help me find her, please? I haven't seen her in days! Who know what could've happened!?
    I'm worried. Will you help me?
        * Yes, of course. This can be a dangerous city.
            Oh no! Do you think something happened to her?
            What should I do?!
                * She's probably fine. She can handle herself.
                    You're right. I'm still worried though.
                        * Don't worry, I'll look for her.
                * Maybe. (stats.chaos += 1)
                    Help me look for her, please! *runs off*
        * I'm sorry, but I don't have time right now. (evil += 1)
            A true friend wouldn't say that.
        * I know where she is! (if bar_mission_solved)
            Really? Where?
                * I saw her on a ship by the docks, it looked like they were ready to set off.
                    Thank you! *runs off*
    ''')
    conversation.start_conversation(convo)
    # conversation.parse_conversation(convo)
    # def input(key):
    #     if key == 'left mouse down' and mouse.hovered_entity in conversation.buttons:
    #         print('add sound here')




    def input(key):
        if key == 'space':
            print(variables.evil)
            # conversation.start_conversation()

    # window.color = color._16
    # window.size = window.fullscreen_size * .5
    Sprite('shore', z=1)
    app.run()



from ursina import *
from ursina.shaders import lit_with_shadows_shader # you have to apply this shader to enties for them to receive shadows.

app = Ursina()
Entity.default_shader = lit_with_shadows_shader
ground = Entity(model='plane', scale=10, texture='grass')
lit_cube = Entity(model='cube', y=1, color=color.light_gray)

light = DirectionalLight()
light.look_at(Vec3(1,-1,1))

shadow_bounds_box = Entity(model='wireframe_cube', scale=5, visible=0)
light.update_bounds(shadow_bounds_box)

EditorCamera(rotation=(30,30,0))
Sky()
app.run()


from ursina import camera, color, mouse, round_to_closest
from ursina.entity import Entity
from ursina.prefabs.button import Button
from ursina.prefabs.input_field import ContentTypes, InputField
from ursina.text import Text
from ursina.vec2 import Vec2


class VecField(Button):
    def __init__(self, default_value=Vec2.zero, character_limit=8, content_type=ContentTypes.math, **kwargs):
        kwargs = dict(parent=camera.ui, scale=(.5,.05), character_limit=character_limit, text='', text_origin=(-.5,0), color=color.black90) | kwargs
        super().__init__(**kwargs)
        if isinstance(default_value, int | float):
            default_value = [default_value, ]

        self.default_value = default_value
        self.fields = []
        self.on_value_changed = None
        w = 1 / len(default_value) *.5

        self.highlight_color = self.color
        self.pressed_color = self.color
        self._dragging_on = None
        self._value_on_drag_start = []
        self._temp_value = []

        self.data_type = float
        if isinstance(default_value[0], int):
            self.data_type = int
            content_type = ContentTypes.int_math

        for i in range(len(default_value)):
            field = InputField(str(self.default_value[i])[:self.character_limit], character_limit=self.character_limit, model='quad', parent=self, limit_content_to=content_type,
                x=(i*w)+w/2, scale=(w,1), z=-1, color=color._8, font=Text.default_monospace_font, submit_on=['enter','tab'])
            field.grid = Entity(parent=field, model='wireframe_quad', color=color.dark_gray, z=-1)
            field.text_field.scale *= .75
            field.on_submit = self.convert_text_to_vector
            # field.on_value_changed = self.convert_text_to_vector
            field.text_field.shortcuts['select_all'] = ('double click', )
            field.text_field.shortcuts['select_word'] = []
            self.fields.append(field)

        self.value = default_value


    def convert_text_to_vector(self):
        vector = []
        for i, field in enumerate(self.fields):
            try:
                # value = type(self.default_value[i])(eval(field.text[:self.character_limit]))
                value = self.data_type(eval(field.text[:self.character_limit]))
                if isinstance(value, self.data_type):
                    vector.append(value)
                    field.text_field.text_entity.text = str(value)[:8]
            except: # invalid/incomplete math
                print('invalid')
                return
        # print('vector:', vector, 'fotype', type(self.default_value))
        if not isinstance(self.default_value, tuple | list):
            vector = type(self.default_value)(*vector)

        self.value = vector


    @property
    def value(self):
        if len(self._value) == 1:
            return self._value[0]
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        if len(value) == 1:
            value = str(value[0])

        if self.on_value_changed is not None:
            self.on_value_changed()

        for i, field in enumerate(self.fields):
            # print('---------', field.text, str(value[i]))
            field.text = str(value[i])[:self.character_limit]


    def input(self, key):
        super().input(key)
        if key == 'left mouse down':
            self._dragging_on = None
            self._temp_value = [e for e in self._value]

            for i, field in enumerate(self.fields):
                if field.hovered:
                    self._dragging_on = field
                    self._value_on_drag_start = self.value
                    return

        elif key == 'left mouse up' and self._dragging_on:
            self._dragging_on = None
            self.value = self._value
            self._temp_value = self._value


    def update(self):
        if mouse.left and self._dragging_on and not self._dragging_on.active:
            # print('drag on:', self._dragging_on, mouse.velocity.x)
            idx = self.fields.index(self._dragging_on)
            if self.data_type is float:
                self._temp_value[idx] += mouse.velocity.x
                self._value = [round_to_closest(e, .01) for e in self._temp_value]
            elif self.data_type is int:
                self._temp_value[idx] += mouse.velocity.x * 10
                self._value = [int(e) for e in self._temp_value]

            for i, field in enumerate(self.fields): # set text temporarily while dragging
                field.text_field.text_entity.text = str(self._value[i])[:self.character_limit]



if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    # Entity(parent=camera.ui, model='quad', scale=.05, z=-10)
    # field = VecField(text=' field', default_value=Vec3(0,0,0))
    def on_value_changed():
        print('set value to:', field.value)

    field = VecField(text=' int list', default_value=[10,1])
    field = VecField(text=' float list', default_value=[1.0,-2.0], y=-.1)
    field = VecField(text=' Vec4', default_value=Vec4(1,-2,0,0), y=-.2)
    field = VecField(text=' float', default_value=1.0, y=-.3)
    field = VecField(text=' int', default_value=0, y=-.4)
    # field.on_value_changed = on_value_changed

    # Entity(color=color.azure, scale=(.5,.05), model='quad', parent=camera.ui)
    # InputField(y=-.1)
    Sprite('shore', color=color.dark_gray)
    app.run()



from ursina import *
from ursina.prefabs.platformer_controller_2d import PlatformerController2d

app = Ursina()
window.color = color.light_gray
camera.orthographic = True
camera.fov = 20
ground = Entity(model='cube', color=color.olive.tint(-.4), z=-.1, y=-1, origin_y=.5, scale=(1000,100,10), collider='box', ignore=True)

random.seed(4)
for i in range(10):
    Entity(model='cube', color=color.dark_gray, collider='box', ignore=True, position=(random.randint(-20,20), random.randint(0,10)), scale=(random.randint(1,20), random.randint(2,5), 10))

player = PlatformerController2d()
player.x = 1
player.y = raycast(player.world_position, player.down).world_point[1] + .01
camera.add_script(SmoothFollow(target=player, offset=[0,5,-30], speed=4))


input_handler.bind('right arrow', 'd')
input_handler.bind('left arrow', 'a')
input_handler.bind('up arrow', 'space')
input_handler.bind('gamepad dpad right', 'd')
input_handler.bind('gamepad dpad left', 'a')
input_handler.bind('gamepad a', 'space')

if application.development_mode:    # cheats
    from ursina.scripts.noclip_mode import NoclipMode2d
    player.add_script(NoclipMode2d())

app.run()



from ursina import camera, color
from ursina.shader import Shader
from ursina.ursinastuff import Func
from ursina.vec2 import Vec2
from ursina.vec3 import Vec3

sky_fade_shader = Shader(name='sky_fade_shader', language=Shader.GLSL, vertex='''
#version 140
uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelViewMatrix;
in vec2 p3d_MultiTexCoord0;
in vec4 p3d_Vertex;
in vec3 p3d_Normal;
in vec4 p3d_Color;
out vec3 world_normal;
out vec3 vertex_world_position;
out vec2 texcoord;
out vec4 vertex_color;

uniform mat4 p3d_ProjectionMatrix;
uniform mat4 p3d_ModelMatrix;

void main() {
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    texcoord = p3d_MultiTexCoord0;

    world_normal = normalize(mat3(p3d_ModelMatrix) * p3d_Normal);
    vertex_world_position = (p3d_ModelMatrix * p3d_Vertex).xyz;
    vertex_color = p3d_Color;

}

''',

fragment='''
#version 150

uniform vec4 p3d_ColorScale;
in vec2 texcoord;
out vec4 fragment_color;

uniform sampler2D p3d_Texture0;
in vec3 world_normal;
in vec3 vertex_world_position;

uniform vec2 texture_scale;

in vec4 vertex_color;
uniform vec3 camera_world_position;

uniform vec4 top_color;
uniform vec4 bottom_color;
uniform vec4 left_color;
uniform vec4 right_color;
uniform vec4 front_color;
uniform vec4 back_color;

uniform sampler2D heightmap;
uniform sampler2D sky_texture;
const float M_PI = 3.14159265358979323846;
uniform float fog_start = 20.;
uniform float fog_end = 200.;


vec4 sky_colored_fog(vec4 color, vec3 vertex_world_position) {
    float distance_to_camera = length(vertex_world_position.xyz - camera_world_position);

    vec3 new_color = color.rgb;

    vec3 vertex_relative_position = vertex_world_position - camera_world_position;
    //vertex_relative_position = vertex_world_position;

    vec2 direction = normalize(vertex_relative_position.zx);
    float angle = atan(direction.y, direction.x); // angle in radians
    float circle_point = clamp(mod((angle / (2.0 * M_PI)) + 1.0, 1.0), 0., 1.);

    vec2 vertical_direction = normalize(vertex_relative_position.yz);
    float vertical_angle = atan(vertical_direction.y, vertical_direction.x); // angle in radians
    float vertical_circle_point = clamp(mod((vertical_angle / (2.0 * M_PI)) + 1.0, 1.0), 0.0, 1.0);

    float y = .1;
    vec4 fog_color = texture(sky_texture, vec2(circle_point, y));
    float offset = .05;
    // fog_color = (
    //     texture(sky_texture, vec2(circle_point, y)).rgb
    //     + texture(sky_texture, vec2(circle_point-offset, y-offset)).rgb
    //     + texture(sky_texture, vec2(circle_point+offset, y-offset)).rgb
    //     + texture(sky_texture, vec2(circle_point-offset, y+offset)).rgb
    //     + texture(sky_texture, vec2(circle_point+offset, y+offset)).rgb
    //     ) / 5.;

    // fog_color = texture(sky_texture, vec2(circle_point, vertical_circle_point)).rgb;
    // float offset = .05;
    // fog_color = (
    //     texture(sky_texture, vec2(circle_point, vertical_circle_point)).rgb
    //     + texture(sky_texture, vec2(circle_point-offset, vertical_circle_point-offset)).rgb
    //     + texture(sky_texture, vec2(circle_point+offset, vertical_circle_point-offset)).rgb
    //     + texture(sky_texture, vec2(circle_point-offset, vertical_circle_point+offset)).rgb
    //     + texture(sky_texture, vec2(circle_point+offset, vertical_circle_point+offset)).rgb
    //     ) / 5.;



    float distance_from_center = length(vertex_relative_position.xz);
    float t = clamp((distance_from_center - fog_start) / (fog_end - fog_start), 0.0, 1.0);
    color.rgb = mix(color.rgb, fog_color.rgb, t * fog_color.a);
    return color;
}

void main(){
    vec4 vert_color = mix(bottom_color, top_color, world_normal.y);
    vert_color *= mix(left_color, right_color, world_normal.x) * 1.5;
    vert_color *= mix(front_color, back_color, world_normal.z) * 2.;
    vert_color = mix(vert_color, vec4(1), 0.1);
    vert_color *= 1.25;

    fragment_color = texture(p3d_Texture0, texcoord) * p3d_ColorScale * vert_color;
    fragment_color = sky_colored_fog(fragment_color, vertex_world_position);
}

''',
default_input = {
    'texture_scale' : Vec2(.1, .1),

    'top_color': color.hsv(220, .12, .82),
    'bottom_color': color.hsv(285, .13, .47),
    'left_color': color.hsv(217, .3, .68),
    'right_color': color.hsv(0, .25, .93),
    'front_color': color.hsv(231, .08, .69),
    'back_color': color.hsv(240, .05, .76),

    'sky_texture': 'sky_default',
    'camera_world_position' : Vec3.zero,
    'fog_start': 50,
    'fog_end': 1000,
    }
,
continuous_input = {
    'camera_world_position': Func(getattr, camera, 'world_position'),
    }
)

if __name__ == '__main__':
    import random

    from ursina import Entity, Sky, Ursina
    from ursina.prefabs.first_person_controller import FirstPersonController
    app = Ursina()
    FirstPersonController(gravity=0)
    # camera.z = -200
    # EditorCamera(rotation_x=60)
    sky = Sky(texture='sky_sunset')
    e = Entity(model='plane', scale=(2000,1,2000), texture='grass', shader=sky_fade_shader)
    e.set_shader_input('sky_texture', sky.texture)
    for i in range(100):
        Entity(model='cube', scale=(25,random.randint(10,150),25), x=-1000+(random.random()*2000), z=-1000+(random.random()*2000), origin_y=-.5, shader=sky_fade_shader, shader_input=dict(sky_texture=sky.texture))
    # camera.fov = 100
    # camera.clip_plane_far = 1000
    print('-----------', camera.clip_plane_far)
    app.run()


from ursina.prefabs.grid_editor import GridEditor
from ursina import color, held_keys, Text, rotate_2d_list
import pyperclip

class ASCIIEditor(GridEditor):
    def __init__(self, size=(61,28), palette=(' ', '#', '|', 'A', '/', '\\', 'o', '_', '-', 'i', 'M', '.'), font=Text.default_monospace_font, canvas_color=color.black, line_height=1.1, **kwargs):
        super().__init__(size=size, palette=palette, canvas_color=canvas_color, **kwargs)
        rotated_grid = rotate_2d_list(self.grid)
        text = '\n'.join([''.join(reversed(line)) for line in reversed(rotated_grid)])

        self.text_entity = Text(parent=self.parent, text=text, x=-.0, y=.5, line_height=line_height, font=font)

        self.scale = (self.text_entity.width, self.text_entity.height)
        self.canvas.scale = 1
        self.text_entity.world_parent = self
        self.text_entity.position = (0, 1, -.001)

    def render(self):
        rotated_grid = rotate_2d_list(self.grid)
        self.text_entity.text = '\n'.join([''.join(reversed(line)) for line in reversed(rotated_grid)])


    def input(self, key):
        super().input(key)
        if held_keys['control'] and key == 'c':
            print(self.text_entity.text)
            pyperclip.copy(self.text_entity.text)
        #
        # if held_keys['control'] and key == 'v' and pyperclip.paste().count('\n') == (h-1):
        #     t.text = pyperclip.paste()
        #     undo_index += 1
        #     undo_stack = undo_stack[:undo_index]
        #     undo_stack.append(deepcopy(grid))

if __name__ == '__main__':
    from ursina import Ursina
    app = Ursina()

    editor = ASCIIEditor()

    app.run()



from ursina.vec2 import Vec2
from ursina.vec3 import Vec3


def project_uvs(model, aspect_ratio=1, direction=Vec3.forward, regenerate=False):
    uvs = list()
    if direction == Vec3.forward:
        for v in model.vertices:
            uvs.append(((v[0]+.5) / aspect_ratio, v[1]+.5))

    elif direction == Vec3.down:
        model.uvs = [Vec2((v[0]+.5) / aspect_ratio, v[2]+.5) for v in model.vertices]

    model.uvs = uvs
    if regenerate:
        model.generate()



if __name__ == '__main__':
    from ursina import EditorCamera, Entity, Ursina
    from ursina.scripts.project_uvs import project_uvs
    app = Ursina()

    e = Entity(model='sphere', texture='ursina_logo')
    project_uvs(e.model)
    EditorCamera()

    app.run()



from ursina import *


app = Ursina()

# Make a simple game so we have something to test with
from ursina.prefabs.first_person_controller import FirstPersonController
player = FirstPersonController(gravity=0, model='cube', color=color.azure)
camera.z = -10
ground = Entity(model='plane', texture='grass', scale=10)

# Create an Entity for handling pausing an unpausing.
# Make sure to set ignore_paused to True so the pause handler itself can still receive input while the game is paused.
pause_handler = Entity(ignore_paused=True)
pause_text = Text('PAUSED', origin=(0,0), scale=2, enabled=False) # Make a Text saying "PAUSED" just to make it clear when it's paused.

def pause_handler_input(key):
    if key == 'escape':
        application.paused = not application.paused # Pause/unpause the game.
        pause_text.enabled = application.paused     # Also toggle "PAUSED" graphic.

pause_handler.input = pause_handler_input   # Assign the input function to the pause handler.


app.run()



from ursina import *
from ursina.models.procedural.plane import Plane


def Cube(bevel=0, subdivisions=(1,1,1), mode='triangle', **kwargs):

    if subdivisions == (1,1,1):
        verts = (
            Vec3(-.5,-.5,-.5), Vec3(.5,-.5,-.5), Vec3(.5,.5,-.5), Vec3(-.5,.5,-.5),
            Vec3(-.5,-.5,.5),  Vec3(.5,-.5,.5),  Vec3(.5,.5,.5), Vec3(-.5,.5,.5)
            )

        tris = (
            (0,1,2,3), (5,4,7,6),   # forward, back
            (3,2,6,7), (4,5,1,0),   # up, down
            (1,5,6,2), (4,0,3,7)    # right, left
            )
        uvs = ((0,0), (1,0), (1,1), (0,1), (0,0), (1,0), (1,1), (0,1))
        cube = Mesh(verts, tris, uvs=uvs, mode=mode, **kwargs)

    else:
        w,h,d = subdivisions
        e = Entity()

        _front =  Entity(parent=e, model=Plane((w,h)), z=-.5, rotation_x=-90)
        _back =   Entity(parent=e, model=Plane((w,h)), z=.5, rotation_x=90)
        _top =    Entity(parent=e, model=Plane((w,d)), y=.5)
        _bottom = Entity(parent=e, model=Plane((w,d)), y=-.5, rotation_x=-180)
        _right =  Entity(parent=e, model=Plane((d,h)), x=.5, rotation_z=90)
        _left =   Entity(parent=e, model=Plane((d,h)), x=-.5, rotation_z=-90)

        cube = e.combine()
        cube.mode = mode
        cube.generate()
        destroy(e)

    return cube


if __name__ == '__main__':
    app = Ursina()

    e = Entity(model=Cube(subdivisions=(3,3,3), mode='line'), color=color.red)
    e = Entity(model=Cube(subdivisions=(1,1,1), mode='line'), color=color.green, x=2)

    _ed = EditorCamera(rotation_speed = 200, panning_speed=200)
    app.run()



from ursina import Shader

pixelation_shader = Shader(
fragment='''
#version 150

uniform sampler2D tex;
in vec2 window_size;
in vec2 uv;
out vec4 color;


void main() {
    float Pixels = 1600.0;
    float dx = 9.0 * (1.0 / Pixels);
    float dy = 16.0 * (1.0 / Pixels);
    vec2 new_uv = vec2(dx * floor(uv.x / dx), dy * floor(uv.y / dy));
    color = texture(tex, new_uv);
}
''')


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='sphere', color=color.orange)
    e = Entity(model='cube', y=-1)
    camera.shader = pixelation_shader
    EditorCamera()

    app.run()



from ursina import *


class SmoothFollow:

    def __init__(self, target=None, offset=Vec3.zero, speed=8, rotation_speed=0, rotation_offset=(0,0,0)):
        self.target = target
        self.offset = offset
        self.speed = speed

        self.rotation_speed = rotation_speed
        self.rotation_offset = rotation_offset


    def update(self):
        if not self.target:
            return

        self.entity.world_position = lerp(
            self.entity.world_position,
            self.target.world_position + Vec3(*self.offset),
            time.dt * self.speed)

        if self.rotation_speed > 0:
            self.entity.world_rotation = lerp(
                self.entity.world_rotation,
                self.target.world_rotation + self.rotation_offset,
                time.dt * self.rotation_speed)




if __name__ == '__main__':
    app = Ursina()

    player = Entity(model='cube', color=color.orange)

    def update():
        player.x += held_keys['d'] * .1
        player.x -= held_keys['a'] * .1

    e = Entity(model='cube')
    sf = e.add_script(SmoothFollow(target=player, offset=(0,2,0)))

    def input(key):
        global sf
        if key == '1' and sf in e.scripts:
            e.scripts.remove(sf)

    EditorCamera()
    app.run()



from ursina import *
from ursina.networking import *
from collections import deque

class InputState:
    def __init__(self):
        self.up = False
        self.down = False
        self.left = False
        self.right = False
        self.sequence_number = 0

    def copy(self):
        cpy = InputState()
        cpy.up = self.up
        cpy.down = self.down
        cpy.left = self.left
        cpy.right = self.right
        cpy.sequence_number = self.sequence_number
        return cpy


def serialize_input_state(writer, input_state):
    writer.write(input_state.up)
    writer.write(input_state.down)
    writer.write(input_state.left)
    writer.write(input_state.right)
    writer.write(input_state.sequence_number)

def deserialize_input_state(reader):
    input_state = InputState()
    input_state.up = reader.read(bool)
    input_state.down = reader.read(bool)
    input_state.left = reader.read(bool)
    input_state.right = reader.read(bool)
    input_state.sequence_number = reader.read(int)
    return input_state

app = Ursina(borderless=False)

start_text = "Press H to host or C to connect."
status_text = Text(text=start_text, origin=(0, 0), z=1)

box = Entity(model="quad", texture="white_cube", color=color.red, parent=camera.ui, scale=0.1, y=0.1)
other_box = Entity(model="quad", texture="white_cube", color=color.blue, parent=camera.ui, scale=0.1, y=-0.1, z=0.5)
other_box.new_x = other_box.x
other_box.new_y = other_box.y
other_box.prev_x = other_box.x
other_box.prev_y = other_box.y

box_speed = 0.4

# Used to show update positions from host without client side prediction.
shadow_box = Entity(model="quad", texture="white_cube", color=color.orange, parent=camera.ui, scale=0.1, y=0.1, z=0.1)

input_state = InputState()
other_input_state = InputState()

inputs_received = deque()
input_buffer = []
send_input_buffer = []

tick_rate = 1.0 / 60.0
tick_timer = 0.0
time_factor = 1.0

update_rate = 1.0 / 20.0
update_timer = 0.0

lerp_time = update_rate * 1.25
lerp_timer = 0.0

input_timer = 0.0

peer = RPCPeer()
peer.register_type(InputState, serialize_input_state, deserialize_input_state)

@rpc(peer)
def set_positions(connection, time_received, my_position: Vec2, other_position: Vec2, sequence_number: int):
    global lerp_timer, time_factor

    if connection.peer.is_hosting():
        return

    # Interpolate other entities (in this case only one).
    other_box.x = other_box.new_x
    other_box.y = other_box.new_y
    other_box.prev_x = other_box.x
    other_box.prev_y = other_box.y
    other_box.new_x = other_position.x
    other_box.new_y = other_position.y
    lerp_timer = 0.0

    sequence_delta = input_state.sequence_number - sequence_number

    # Maybe slow down if ahead of host.
    max_delta = ((update_rate / tick_rate) + 1) * 2
    if sequence_delta > max_delta:
        time_factor = 0.95
    elif sequence_delta < max_delta * 0.75:
        time_factor = 1.0

    # Reconcile with host.
    box.x = my_position.x
    box.y = my_position.y
    shadow_box.x = box.x
    shadow_box.y = box.y
    if sequence_delta > 0 and sequence_delta < len(input_buffer):
        # Re-apply all inputs after the last processed input.
        for state in input_buffer[len(input_buffer) - sequence_delta:]:
            box.x += float(int(state.right) - int(state.left)) * box_speed * tick_rate
            box.y += float(int(state.up) - int(state.down)) * box_speed * tick_rate


@rpc(peer)
def set_inputs(connection, time_received, input_states: list[InputState]):
    if not connection.peer.is_hosting():
        return

    if len(inputs_received) > 100:
        # Host is being spammed, disconnect.
        print("Peer is spamming inputs, disconnecting...")
        connection.disconnect()

    for state in input_states:
        inputs_received.append(state)

def tick(dt):
    global last_input_sequence_number_processed

    if time_factor < 1.0:
        print("Host is lagging, slowing down local simulation.")

    input_state.up = bool(held_keys["w"])
    input_state.down = bool(held_keys["s"])
    input_state.right = bool(held_keys["d"])
    input_state.left = bool(held_keys["a"])
    box.x += float(int(input_state.right) - int(input_state.left)) * box_speed * dt
    box.y += float(int(input_state.up) - int(input_state.down)) * box_speed * dt

    if not peer.is_hosting():
        input_state.sequence_number += 1
        input_buffer.append(input_state.copy())
        if len(input_buffer) >= 100:
            input_buffer.pop(0)
        send_input_buffer.append(input_buffer[-1])
        if len(send_input_buffer) > 10:
            send_input_buffer.pop(0)
    else:
        last_input_sequence_number_processed = other_input_state.sequence_number
        if len(inputs_received) > 0:
            state = inputs_received.popleft()
            other_input_state.up = state.up
            other_input_state.down = state.down
            other_input_state.left = state.left
            other_input_state.right = state.right
            other_input_state.sequence_number = state.sequence_number
        else:
            other_input_state.up = False
            other_input_state.down = False
            other_input_state.left = False
            other_input_state.right = False
        other_box.x += float(int(other_input_state.right) - int(other_input_state.left)) * box_speed * dt
        other_box.y += float(int(other_input_state.up) - int(other_input_state.down)) * box_speed * dt

def update():
    global update_timer, lerp_timer, tick_timer

    peer.update()
    if not peer.is_running():
        status_text.text = start_text
        box.x = 0.0
        box.y = 0.1
        other_box.x = 0.0
        other_box.y = -0.1
        return

    if peer.is_hosting():
        status_text.text = "Hosting on localhost, port 8080.\nWASD to move."
    else:
        status_text.text = "Connected to host with address localhost, port 8080.\nWASD to move."

    tick_timer += time.dt * time_factor
    while tick_timer >= tick_rate:
        tick(tick_rate)
        tick_timer -= tick_rate

    if not peer.is_hosting():
        lerp_timer += time.dt
        other_box.x = lerp(other_box.prev_x, other_box.new_x, lerp_timer / lerp_time)
        other_box.y = lerp(other_box.prev_y, other_box.new_y, lerp_timer / lerp_time)
        if lerp_timer >= lerp_time:
            other_box.x = other_box.new_x
            other_box.y = other_box.new_y
            other_box.prev_x = other_box.x
            other_box.prev_y = other_box.y
            other_box.new_x = other_box.x
            other_box.new_y = other_box.y
            lerp_timer = 0.0

    update_timer += time.dt
    if update_timer >= update_rate:
        if peer.is_running() and peer.connection_count() > 0:
            if peer.is_hosting():
                peer.set_positions(
                    peer.get_connections()[0],
                    Vec2(other_box.x, other_box.y),
                    Vec2(box.x, box.y),
                    other_input_state.sequence_number
                )
            else:
                peer.set_inputs(peer.get_connections()[0], send_input_buffer)
                send_input_buffer.clear()
        update_timer = 0.0

def input(key):
    if key == "h":
        box.y = 0.1
        other_box.y = -0.1
        other_box.new_x = other_box.x
        other_box.new_y = other_box.y
        other_box.prev_x = other_box.x
        other_box.prev_y = other_box.y
        shadow_box.visible = False

        peer.start("localhost", 8080, is_host=True)
    elif key == "c":
        box.y = -0.1
        other_box.y = 0.1
        other_box.new_x = other_box.x
        other_box.new_y = other_box.y
        other_box.prev_x = other_box.x
        other_box.prev_y = other_box.y
        shadow_box.visible = True

        peer.start("localhost", 8080, is_host=False)

app.run()



from ursina import *

'''
nineslice mesh:

11-10--15--14
|   |   |   |
8---9--12--13
|   |   |   |
3---2---7---6
|   |   |   |
0---1---4---5

'''


class NineSlice(Mesh):
    def __init__(self, entity_scale=Vec3.one, scale_multiplier=1):
        verts = [Vec3(*e,0)*scale_multiplier for e in (
            (-.5,-.5), (0,-.5), (0,0), (-.5,0),
            (0,-.5), (.5,-.5), (.5,-0), (0,-0),
            (-.5,0), (-0,0), (-0,.5), (-.5,.5),
            (0,0), (.5,0), (.5,.5), (0,.5),
            )]

        offset = (entity_scale - Vec3.one)
        for i in range(0,4):
            verts[i] -= offset * .5
            verts[i] *= Vec3.one / entity_scale
        for i in range(4,8):
            verts[i] += offset * Vec3(1,-1,1) * .5
            verts[i] *= Vec3.one / entity_scale
        for i in range(8,11):
            verts[i] += offset * Vec3(-1,1,1) * .5
            verts[i] *= Vec3.one / entity_scale
        for i in range(12,16):
            verts[i] += offset * .5
            verts[i] *= Vec3.one / entity_scale

        super().__init__(
            vertices=verts,
            triangles=(
                (0,1,2,3),
                (1,4,7,2),
                (4,5,6,7),
                (3,2,9,8,),
                (2,7,12,9),
                (7,6,13,12),
                (8,9,10,11),
                (9,12,15,10),
                (12,13,14,15)
                ),
            uvs=(
                (0,0), (.5,0), (.5,.5), (0,.5),
                (.5,0), (1,0), (1,.5), (.5,.5),
                (0,.5), (.5,.5), (.5,1), (0,1),
                (.5,.5), (1,.5), (1,1), (.5,1)
                ),
            )



if __name__ == '__main__':
    app = Ursina()
    m = NineSlice(entity_scale=Vec3(2,1,1),
        # scale_multiplier=1.4,   # account for whitespace and drop shadow in the texture
        )

    window.color = color._240

    Entity(parent=camera.ui, model='wireframe_quad', scale=Vec2(2,1)*.1, color=color.red)
    for i, (nineslice_texture, scale_multiplier) in enumerate(
        zip(
            ('nineslice_rainbow', 'nineslice_double', 'circle'),
            (1.4, 1.4, 1), strict=True
        )):
        b = Draggable(
            model=NineSlice(entity_scale=Vec3(2,1,1), scale_multiplier=scale_multiplier),
            scale=Vec2(2,1)*.1,
            texture=nineslice_texture,
            text='Next',
            color=color.white,
            text_color=color.dark_gray,
            text_origin=(0,0),
            highlight_color=hsv(210,.1,1),
            pressed_scale=.9,
            pressed_color=hsv(180,.2,.7),
            # wireframe=True,
            y=-i*.2
            )

    # button1 = Entity(text="Hi !!", model=NineSlice(0.2, 0.3), position=(-0.2, -0.3), texture="nine_sliced", color=color.white)
    # button2 = Entity(text="Hello !!!", model=NineSlice(0.5, 0.3), position=(0.2, -0.3), texture="nine_sliced", color=color.white)
    # button3 = Entity(text="As you can see, the texture goes \nwell for every button size.", model=NineSlice(0.4, 0.4), position=(0, 0.1), texture="nine_sliced", color=color.white)
    EditorCamera()
    app.run()


from ursina import color
from ursina.shader import Shader

matcap_shader = Shader(name='matcap_shader', language=Shader.GLSL, vertex = '''#version 140
uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelMatrix;
uniform mat4 p3d_ModelViewMatrix;
uniform mat3 p3d_NormalMatrix;
in vec4 p3d_Vertex;
in vec3 p3d_Normal;

out vec3 eye;
out vec3 view_normal;
// reflect alternative:
// r = e - 2. * dot( n, e ) * n;

void main() {
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;

    eye = normalize(vec3(p3d_ModelViewMatrix * vec4(p3d_Vertex.xyz, 1.0)));
    view_normal = normalize( p3d_NormalMatrix * p3d_Normal );
}
''',
fragment='''
#version 130
uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;

in vec3 eye;
in vec3 view_normal;
out vec4 fragColor;

void main() {
    vec3 r = reflect(eye, view_normal);
    float m = 2. * sqrt(pow(r.x, 2.) + pow(r.y, 2.) + pow(r.z + 1., 2.));
    vec2 vN = r.xy / m + .5;

    fragColor = texture(p3d_Texture0, vN) * p3d_ColorScale;
}

''',
)



if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    app = Ursina()
    window.color=color.black

    shader = matcap_shader

    a = WhiteCube(shader=shader, texture='shore')
    b = WhiteSphere(shader=shader, rotation_y=180, x=3, texture='shore')
    # AzureSphere(shader=a.shader, y=2)
    GrayPlane(scale=10, y=-2, texture='shore')

    Sky(color=color.light_gray)
    EditorCamera()

    def update():
        b.rotation_z += 1
        b.rotation_y += 1
        b.rotation_x += 1
    # EditorCamera()
    print('-----------------', repr(a.shader))
    app.run()



from ursina import *
from ursina.prefabs.grid_editor import GridEditor
import re

class Tilemap(GridEditor):
    def __init__(self, tilemap='', tileset='', tileset_size=(8,8), **kwargs):
        if isinstance(tilemap, str):
            self.tilemap = load_texture(tilemap)

        self.grid = self.tilemap.pixels
        super().__init__(texture=self.tilemap, size=self.tilemap.size, palette=(color.white, color.black, color.green, color.blue, color.red), edit_mode=False, **kwargs)

        # self.canvas.visible_self = False
        self.canvas.alpha = .5
        self.tileset = tileset
        self._tileset_size = tileset_size
        self.model = Mesh()
        self.texture = tileset
        self.colliders = []

        self.auto_render = False
        self.outline = Entity(parent=self, model=Quad(segments=0, mode='line', thickness=1), color=color.cyan, z=.01, origin=(-.5,-.5), enabled=self.edit_mode)

        self._quad = Quad(segments=0)
        self._quad.vertices = [Vec3(*v)+Vec3(.5,.5,0) for v in self._quad.vertices]
        self._garbage = Entity(parent=self.canvas, add_to_scene_entities=False)

        self.uv_dict = {
            '11111111' : [(4,1), (5,1), (6,1), (7,1)],     # fill

            '0.11111.' : [(1,3), (2,3), (5,3), (6,3)],     # top
            '1.0.1111' : [(3,2), (3,1)],     # right
            '111.0.11' : [(2,0), (1,0), (6,2), (5,2)],     # bot
            '11111.0.' : [(0,1), (0,2)],     # left

            '000.111.' : [(3,3), ],     # corner_top_right
            '1.000.11' : [(3,0), ],     # corner_bot_right
            '111.000.' : [(0,0), ],     # corner_bot_left
            '0.111.00' : [(0,3), ],     # corner_top_left

            '10111111' : [(1,1), ],     #inner_corner_bot_left
            '11101111' : [(1,2), ],     #inner_corner_top_left
            '11111011' : [(2,2), ],     #inner_corner_top_right
            '11111110' : [(2,1), ],     #inner_corner_bot_right

        }
        self.single_block_coordinates = [(4,0), (5,0), (6,0), (7,0)]
        self.variation_chance = [0,0,0,0,1,1,1,2,2,3]

        if 'min' in self.texture.name:
            self.uv_dict = {
                '11111111' : [(1,1)],     # fill

                '0.11111.' : [(1,2)],     # top
                '111.0.11' : [(1,0), ],     # bot
                '1.0.1111' : [(0,1), '-1,1'],     # right
                '11111.0.' : [(0,1)],     # left

                '0.111.00' : [(0,2), ],     # corner_top_left
                '000.111.' : [(0,2), '-1,1'],     # corner_top_right
                '1.000.11' : [(0,2), '-1,-1'],     # corner_bot_right
                '111.000.' : [(0,2), '1,-1'],     # corner_bot_left

                '11111110' : [(2,0), ],     #inner_corner_bot_right
                '10111111' : [(2,0), '-1,1'],     #inner_corner_bot_left
                '11111011' : [(2,1), ],     #inner_corner_top_right
                '11101111' : [(2,1), '-1,1'],     #inner_corner_top_left
            }
            self.single_block_coordinates = [(2,2)]
            self.variation_chance = [0,]

        self.uv_margin = .002
        self.render()


    def update(self):
        if not self.edit_mode:
            return

        super().update()
        if mouse.left:
            self.draw_temp(self.cursor.position)


    def draw_temp(self, position):
        e = Entity(
            parent=self._garbage,
            model='quad',
            scale=Vec3(1/self.tilemap.width, 1/self.tilemap.height, 1) * self.brush_size,
            position=Vec3(self.cursor.x/self.tilemap.width, self.cursor.y/self.tilemap.height, -.1),
            # always_on_top=True,
            texture=self.texture,
            texture_scale=Vec2.one/self._tileset_size,
            texture_offset=Vec2(.5, .0),
            origin=(-.5,-.5),
            ignore=True,
            )
        if self.selected_char == self.palette[0]:
            e.color = window.color
            e.texture = None


    def input(self, key):
        super().input(key)
        if key == 'left mouse up':
            for e in self._garbage.children:
                destroy(e)


    def render(self):
        self.scale = self.tilemap.size
        self.model.clear()

        for e in self.colliders:
            destroy(e)
        self.colliders.clear()

        tile_size = Vec2.one/self._tileset_size

        i = 0
        for y in range(self.tilemap.height):

            collider = None
            for x in range(self.tilemap.width):
                col = self.grid[x][y]

                if col == color.white and collider: # end collider
                    collider = None

                if col != color.white: #
                    self.model.vertices.extend([Vec3(x/self.tilemap.width, y/self.tilemap.height, 0) + (v*1/self.tilemap.width) for v in self._quad.vertices]) # add quad vertices, but offset.
                    self.model.triangles.append([i+j for j in range(4)])

                    neighbours = list()
                    # register neighbours clockwise starting from the top
                    for offset in [(0,1), (1,1), (1,0), (1,-1), (0,-1), (-1,-1), (-1,0), (-1,1)]:
                        try:
                            neighbours.append(self.grid[x+offset[0]][y+offset[1]] != color.white)
                        except IndexError:
                            neighbours.append(1)

                    neighbours = ''.join([str(int(e)) for e in neighbours])

                    # if collider is None and neighbours != '11111111':
                    #     collider = Entity(
                    #         parent=self,
                    #         model='cube',
                    #         scale_x=Vec3(0, 1/self.tilemap.height, 1),
                    #         origin_x=-.5,
                    #         position=(x-.5, y/self.tilemap.height),
                    #         collider='box',
                    #         color=color.red
                    #         )
                    #     self.colliders.append(collider)
                    # if collider and neighbours == '11111111': # end collider if it's a middle block
                    #     collider = None
                    # if collider:
                    #     collider.scale_x += 1/self.tilemap.width

                    random.seed(y+x)
                    variation_index = random.choice(self.variation_chance)
                    tile_scale = '1,1'

                    for key, value in self.uv_dict.items():
                        if re.match(key, neighbours):
                            if isinstance(value[-1], str):
                                tile_scale = value[-1]

                            variation_index = min(variation_index, len(value)-1-int(tile_scale=='1,1'))
                            _x, _y = value[variation_index]
                            break
                        else:
                            _x, _y = self.single_block_coordinates[variation_index]

                    uv = [
                        (tile_size * Vec2(_x,  _y))   + Vec2( self.uv_margin,  self.uv_margin),
                        (tile_size * Vec2(_x+1,_y))   + Vec2(-self.uv_margin,  self.uv_margin),
                        (tile_size * Vec2(_x+1,_y+1)) + Vec2(-self.uv_margin, -self.uv_margin),
                        (tile_size * Vec2(_x,  _y+1)) + Vec2( self.uv_margin, -self.uv_margin),
                    ]
                    print(uv)
                    if tile_scale == '1,1':
                        pass
                    elif tile_scale == '-1,1':
                        a, b, c, d = uv
                        uv = [b, a, d, c]
                    elif tile_scale == '1,-1':
                        a, b, c, d = uv
                        uv = [d, c, b, a]
                    elif tile_scale == '-1,-1':
                        a, b, c, d = uv
                        uv = [c, d, a, b]


                    self.model.uvs.extend(uv)
                    i += 4


        self.model.generate() # call to create the mesh


    def save(self):
        for y in range(self.tilemap.height):
            for x in range(self.tilemap.width):
                self.tilemap.set_pixel(x, y, self.grid[x][y])

        if self.tilemap.path:
            self.tilemap.save(self.tilemap.path)
            print('saving:', self.tilemap.path)



if __name__ == '__main__':
    app = Ursina()
    EditorCamera(rotation_speed=0)
    tilemap = Tilemap('tilemap_test_level', tileset='test_tileset', tileset_size=Vec2(8,4), parent=scene)
    tilemap.canvas.texture = 'tilemap_test_level'
    # tilemap = Tilemap('brick', tileset='tileset_cave', tileset_size=(8,4), parent=scene)
    camera.orthographic = True
    camera.position = tilemap.tilemap.size / 2
    camera.fov = tilemap.tilemap.height

    Text('press tab to toggle edit mode', position=window.top_left)

    app.run()



from ursina import *


helper = None


def terraincast(world_position, terrain_entity, height_values=None, return_normals=False):    # uses x and z to return y on terrain.
    global helper
    if not helper:
        helper = Entity()

    helper.parent = terrain_entity.model
    helper.world_position = world_position

    pos = helper.get_position(relative_to=terrain_entity.model) + Vec3(.5,0,.5)

    if height_values is None:
        height_values = terrain_entity.model.height_values

    w, d = len(height_values), len(height_values[0])

    if pos.x >= 0 and pos.x < 1 and pos.z >= 0 and pos.z < 1:
        pos *= Vec3(w-1, 0, d-1)
        helper.y = height_values[int(pos.x)][int(pos.z)]
        x, _, z = pos

        point = height_values[int(floor(x))][int(floor(z))]
        normal = Vec3(0,0,0)

        if ceil(x) - x > 0 and ceil(z) - z > 0:
            point_e =  height_values[int(min(w-1, ceil(x)))][int(floor(z))]
            point_n =  height_values[int(floor(x))][int(min(d-1, ceil(z)))]
            point_ne = height_values[int(min(w-1, ceil(x)))][int(min(d-1, ceil(z)))]

            u0v0 = point * (ceil(x) - x) * (ceil(z) - z) # interpolated (x0, z0)
            u1v0 = point_e * (x - floor(x)) * (ceil(z) - z) # interpolated (x1, z0)
            u0v1 = point_n * (ceil(x) - x) * (z - floor(z)) # interpolated (x0, z1)
            u1v1 = point_ne * (x - floor(x)) * (z - floor(z)) # interpolated (x1, z1)

            point = u0v0 + u1v0 + u0v1 + u1v1  #estimate
            # normal = Vec3(2*(point_e-L), 2*(B-T), -4).Normalize();
            # print(point_e)
            if return_normals:
                normal = Vec3(0,1,0)
                p1 = Vec3(min(w-1, ceil(x)), point_e, floor(z))
                p2 = Vec3(floor(x), point, floor(z))
                p3 = Vec3(floor(x), point_n, min(d-1, ceil(z)))
                normal = (p2 - p1).cross(p3-p1).normalized()
                # print(normal)


        helper.y = point * terrain_entity.scale_y / 255
        if not return_normals:
            return helper.y

        return helper.y, normal

    if not return_normals:
        return None

    return None, None


if __name__ == '__main__':
    app = Ursina()

    terrain_entity = Entity(model=Terrain('heightmap_1', skip=8), scale=(40, 5, 20), texture='heightmap_1')
    player = Entity(model='sphere', color=color.azure, scale=.2, origin_y=-.5)


    hv = terrain_entity.model.height_values

    def update():
        direction = Vec3(held_keys['d'] - held_keys['a'], 0, held_keys['w'] - held_keys['s']).normalized()
        player.position += direction * time.dt * 4
        y = terraincast(player.world_position, terrain_entity, hv)
        if y is not None:
            player.y = y
        # test.world_position = player.world_position
        # test.look_at(test.world_position + normal)

    EditorCamera()
    # from ursina.shaders import ssao_shader
    # camera.shader = ssao_shader
    Sky()

    app.run()



from ursina import *                    # this will import everything we need from ursina with just one line.

class Player(Entity):                   # inherits Entity, ursina's 'god class'
    def __init__(self):
        super().__init__()
        self.model = 'cube'             # finds a 3d model by name
        self.color = color.orange
        self.scale_y = 2

    def update(self):               # because Player is an Entity, update gets automatically called by the engine.
        self.x += held_keys['d'] * 1/128
        self.x -= held_keys['a'] * 1/128


app = Ursina()
s = Sprite(texture='beach_level_pattern')
camera.orthographic = True
camera.fov = 1

player = Player()
player.scale *= 1/128 * 16
camera.world_parent = player
app.run()                               # opens a window and starts the game.



from ursina import *
from ursina.shaders import lit_with_shadows_shader
random.seed(0)

app = Ursina()

Entity(model='cube', y=150, color=color.red)

# Normally I'd load a heightmap texture, but for this example I'll include a heightmap texture in the script itself, as base64.
data = '''data:image/png;base64, '''
heightmap = Texture(data)

terrain_model = Terrain(heightmap)
Entity.default_shader = lit_with_shadows_shader
# print(terrain_model.height_values.tolist())
terrain = Entity(model=terrain_model, texture='grass', texture_scale=(3,3), scale=128)
# Instead of using a texture on the terrain, map them to vertex colors.
# Later, we'll use these colors to tint the grass we place, so the grass match the ground.
for uv in terrain.model.uvs:
    x = uv[0] * terrain.texture_scale[0]
    y = uv[1] * terrain.texture_scale[1]
    x -= int(x)
    y -= int(y)
    x = int(x * terrain.texture.width)
    y = int(y * terrain.texture.height)
    terrain.model.colors.append(terrain.texture.get_pixel(x, y).tint(-.25))

terrain.model.generate()
terrain.texture = None

scene.fog_density = .003


# Use two quads crossing each other to create a grass model
quad_model = load_model('quad', use_deepcopy=True)
grass_models = [deepcopy(quad_model) for i in range(4)]
from ursina.ursinamath import rotate_around_point_2d
for i, mesh in enumerate(grass_models):
    p = Entity()
    Entity(parent=p, model='quad', rotation_y=i*(360/5), origin_y=-.5, scale=1.5)
    Entity(parent=p, model='quad', rotation_y=(i*(360/5))+90, origin_y=-.5, scale=1.5)
    p.combine()
    grass_models[i] = p.model

# Instead of generating blue noise from scratch, which could be slow, use a pre-generated looping pattern to get the same effect
blue_noise = Empty(
    width = 64,
    height = 64,
    pattern = [(40, 1), (47, 1), (22, 2), (8, 3), (16, 3), (58, 4), (29, 6), (51, 6), (38, 7), (45, 8), (15, 9), (4, 10), (23, 10), (34, 12), (9, 13), (53, 13), (60, 13), (42, 14), (19, 16), (3, 17), (31, 17), (51, 18), (12, 19), (24, 19), (38, 19), (45, 21), (61, 21), (6, 23), (16, 24), (29, 24), (55, 25), (36, 26), (21, 27), (49, 27), (42, 28), (61, 28), (4, 29), (12, 30), (33, 31), (25, 32), (52, 33), (2, 34), (45, 34), (59, 34), (38, 35), (11, 37), (28, 37), (19, 38), (4, 39), (34, 39), (48, 40), (54, 41), (61, 41), (22, 44), (29, 44), (8, 45), (37, 45), (42, 45), (14, 46), (1, 47), (51, 47), (58, 50), (25, 51), (32, 51), (46, 51), (5, 52), (19, 52), (12, 54), (38, 54), (54, 55), (28, 57), (62, 57), (6, 58), (41, 58), (48, 58), (20, 59), (13, 61), (56, 62)]
)

perlin_noise_pixels = load_texture('perlin_noise').pixels


density = .8
grass_parent = Entity(texture='grass_lump', double_sided=True, model='grass_combined')
grass_colors = list(set([colr.tint(.2) for colr in terrain.model.colors]))

def place_grass():
    for __X in range(-8,8,1):
        for __Z in range(-8,8,1):

            for pos in blue_noise.pattern:
                if random.random() < 1-density:
                    continue

                x = ((__X * blue_noise.width)  + pos[0]) * .125
                z = ((__Z * blue_noise.height) + pos[1]) * .125
                y, n = terraincast(Vec3(x,300,z), terrain, return_normals=True)
                if not y:
                    continue

                # if sample_bilinear(perlin_noise_pixels, x, y, clamp_if_outside=False) < .5:
                #     continue

                grass_parent.model.vertices.extend([(v*1)+Vec3(x,y,z) for v in random.choice(grass_models).vertices])

                random_color = random.choice(grass_colors)
                # random_color = color.white

                grass_parent.model.colors.extend([random_color, ] * len(grass_models[0].vertices))
                grass_parent.model.uvs.extend(grass_models[0].uvs)

    grass_parent.model.generate()
    # Since placing the grass can be slow, save the result to disk and load it instead of generating it again next time.
    grass_parent.model.save('grass_combined.bam')

if not grass_parent.model:
    grass_parent.model = Mesh()
    place_grass()


player = EditorCamera(model=Capsule(height=1.5), origin_y=-.75, color=color.orange)
camera.fov = 100
player_speed = 8
def update():
    player.position += player.forward * (held_keys['w'] - held_keys['s']) * time.dt * player_speed
    player.position += player.right * (held_keys['d'] - held_keys['a']) * time.dt * player_speed
    player.y = terraincast(player.world_position+Vec3(0,100,0), terrain)



def input(key):
    if key == 'space':
        grass_parent.model = Mesh()
        place_grass()

Sky()
DirectionalLight().look_at(Vec3(1,-1,1))
# window.size = Vec2(1920,1080)
# window.borderless = True
# window.editor_ui.enabled = False

app.run()



from ursina import *


app = Ursina()

start_point = Draggable(model='circle', color=color.orange, scale=.025, position=(-0,-0))
end_point = Draggable(model='circle', color=color.orange, scale=.025, position=(.2,.1))

line = Entity(parent=camera.ui, model='line', origin_x=-.5) # set origin_x = -.5 o it starts at the left and not the middle.

def update():
    line.position = start_point.position
    line.look_at_2d(end_point)  # could also ise .look_at() (3d) for this.
    line.rotation_z -= 90 # look_at_2d assumes up as forward, so offset by -90 degrees.
    line.scale_x = distance_2d(start_point, end_point)

app.run()



from ursina.scripts.property_generator import generate_properties_for_class

if __name__ == '__main__':
    from ursina.ursinastuff import _test


def flatten_list(target_list):
    import itertools
    return list(itertools.chain(*target_list))


def flatten_completely(target_list):
    for i in target_list:
        if isinstance(i, tuple | list):
            yield from flatten_list(i)
        else:
            yield i


def enumerate_2d(lst):    # usage: for (x, y), value in enumerate_2d(my_2d_list)
    if isinstance(lst, Array2D):
        width = lst.width
        height = lst.height
    else:
        width = len(lst)
        height = len(lst[0])
    for y in range(height):
        for x in range(width):
            yield (x, y), lst[x][y]


def enumerate_3d(target_3d_list):   # usage: for (x, y, z), value in enumerate_3d(my_3d_list)
    for x, vertical_slice in enumerate(target_3d_list):
        for y, log in enumerate(vertical_slice):
            for z, value in enumerate(log):
                yield (x, y, z), value



@generate_properties_for_class()
class Array2D(list):
    __slots__ = ('width', 'height', 'default_value')

    def __init__(self, width:int=None, height:int=None, default_value=0, data:list[list]=None):
        self.default_value = default_value

        if data is not None:                # initialize with provided data
            self.width = len(data)
            self.height = len(data[0])
            if self.width == 0 or self.height == 0:
                raise ValueError("width and height must be >= 1.")

            if any(len(row) != self.height for row in data):
                raise ValueError("all rows in the data must have the same length.")

            super().__init__(data)

        else:   # Initialize with default values
            if width is None or height is None:
                raise ValueError("width and height must be provided if no data is given.")

            self.width = int(width)
            self.height = int(height)
            super().__init__([[self.default_value for _ in range(self.height)] for _ in range(self.width)])

    @staticmethod
    def from_string(string, convert_to_type=str, starts_lower_left=True):
            string = string.strip()
            if starts_lower_left:
                data = [[convert_to_type(word.strip()) for word in line.split(',') if word] for line in string.split('\n') if line]
                data = [list(row) for row in zip(*data[::-1], strict=True)]  # rotate
                return Array2D(data=data)
            else:
                return Array2D(data=[[word.strip() for word in line.split(',') if word] for line in string.split('\n') if line])

    def to_string(self, separator=', ', always_separate=False):
        lines = []
        longest_width = max(len(str(value)) for (_,_), value in enumerate_2d(self))

        if always_separate or longest_width > 1:    # separate each element with ', '
            for y in range(self.height-1, -1, -1):
                line = separator.join([f'{str(self[x][y]):>{longest_width}}' for x in range(self.width)])
                lines.append(line)
        else:   # make string without spaces since all the values are the same width
            for y in range(self.height-1, -1, -1):
                line = ''.join([str(self[x][y]) for x in range(self.width)])
                lines.append(line)


        return '\n'.join(lines)

    @property
    def size(self):
        from ursina.vec2 import Vec2
        return Vec2(self.width, self.height)

    def rows_getter(self):
        return [[self[x][y] for x in range(self.width)] for y in range(self.height)]

    def columns_getter(self):
        return [[self[x][y] for y in range(self.height)] for x in range(self.width)]

    def __str__(self):
        lines = self.to_string().strip().split('\n')
        longest_number = int(len(str(self.height)))
        for y in range(self.height-1, -1, -1):
            lines[y] = f'{self.height-1-y:<{longest_number}}| {lines[y]}'

        padding = longest_number + 2
        lines.append(f'{"o":<{padding}}{"-"*(self.width)}w:{self.width}')
        result = '\n'+'\n'.join(lines)
        return result


    def map(self, func):
        for (x, y), value in enumerate_2d(self):
            self[x][y] = func(x, y, value)
        return self


    def reset(self):
        for x in range(self.width):
            for y in range(self.height):
                self[x][y] = self.default_value

    def paste(self, data, x, y, ignore=-1):
        for true_x in range(x, min(self.width, x+data.width)):
            for true_y in range(y, min(self.height, y+data.height)):
                if data[true_x-x][true_y-y] == ignore:
                    continue
                self[true_x][true_y] = data[true_x-x][true_y-y]


    def add_margin(self, top=0, right=0, bottom=0, left=0, value=None):
        if top < 0 or right < 0 or right < 0 or left < 0:
            raise ValueError('input must be >= 0')
        if value is None:
            value = self.default_value

        new_width = self.width + right + left
        new_height = self.height + top + bottom
        new_array = Array2D(new_width, new_height, default_value=value)
        new_array.paste(self, left, bottom)
        return new_array

    @property
    def bounds(self):
        from ursina.ursinamath import Bounds
        from ursina.vec3 import Vec3
        min_x = self.width
        min_y = self.height
        max_x = 0
        max_y = 0

        for (x, y), value in enumerate_2d(self):
            if value:
                min_x = min(min_x, x)
                min_y = min(min_y, y)
                max_x = max(max_x, x)
                max_y = max(max_y, y)

        return Bounds(start=Vec3(min_x, min_y, 1), end=Vec3(max_x, max_y, 1))

    def get(self, x, y, default=0):
        x, y = int(x), int(y)
        if x < 0 or x >= self.width or y < 0 or y >= self.height:
            return default
        return self[x][y]


    def set(self, x, y, value):
        x, y = int(x), int(y)
        if x < 0 or x >= self.width or y < 0 or y >= self.height:
            return
        self[x][y] = value


    def get_area(self, start, end, allow_out_of_bounds=False):
        cropped_array = Array2D(width=end[0]-start[0], height=end[1]-start[1], default_value=self.default_value)
        # print('original_size:', self.width, self.height, 'new_size:', end[0]-start[0], end[1]-start[1])
        if not allow_out_of_bounds:
            for (x, y), _ in enumerate_2d(cropped_array):
                cropped_array[x][y] = self[int(x+start[0])][int(y+start[1])]
        else:
            for (x, y), _ in enumerate_2d(cropped_array):
                cropped_array[x][y] = self.get(x+start[0], y+start[1], default=self.default_value)

        return cropped_array



if __name__ == '__main__':
    from textwrap import dedent, indent
    grid = Array2D(width=16, height=8)
    # print(grid)
    padded_grid = grid.add_margin(top=4, right=7, bottom=3, left=2, value=7)
    # print('added margin:', padded_grid)
    # print('cropped_array:\n', padded_grid.get_area((2,3), (padded_grid.width-7, padded_grid.height-4)))

    _test(
        Array2D(data=[[1,6], [2,7], [3,8], [4,9], [5,10]]).to_string()
        ==
        indent(dedent('''
        6,  7,  8,  9, 10
        1,  2,  3,  4,  5
        ''').strip(), ' ')
    )

    _test(Array2D(data=[[1,6], [2,7], [3,8], [4,9], [5,10]]).rows == [[1,2,3,4,5], [6,7,8,9,10]])

    _test(Array2D.from_string('''
        0,1,0
        2,3,4
        0,1,1
        0,0,0
        ''', int) ==
        Array2D(data=[
        [0,0,2,0],
        [0,1,3,1],
        [0,1,4,0]
        ])
    )


class Array3D(list):
    __slots__ = ('width', 'height', 'depth', 'default_value')

    def __init__(self, width:int=None, height:int=None, depth:int=None, default_value=0, data:list[list]=None):
        self.default_value = default_value

        if data is not None:                # initialize with provided data
            self.width = len(data)
            self.height = len(data[0])
            self.depth = len(data[0][0])
            if self.width == 0 or self.height == 0 or self.depth == 0:
                raise ValueError("width, height, depth must be >= 1.")

            if any(len(row) != self.height for row in data):
                raise ValueError("all rows in the data must have the same length.")

            super().__init__(data)

        else:   # Initialize with default values
            if width is None or height is None or depth is None:
                raise ValueError("width and height and depth must be provided if no data is given.")

            self.width = int(width)
            self.height = int(height)
            self.depth = int(depth)
            super().__init__([[[self.default_value for _ in range(self.depth)] for _ in range(self.height)] for _ in range(self.width)])


    @property
    def size(self):
        from ursina.vec3 import Vec3
        return Vec3(self.width, self.height, self.depth)


    @property
    def bounds(self):
        from ursina.ursinamath import Bounds
        from ursina.vec3 import Vec3
        min_x = self.width
        min_y = self.height
        min_z = self.depth
        max_x = 0
        max_y = 0
        max_z = 0

        for (x, y, z), value in enumerate_3d(self):
            if value:
                min_x = min(min_x, x)
                min_y = min(min_y, y)
                min_z = min(min_z, z)
                max_x = max(max_x, x)
                max_y = max(max_y, y)
                max_z = max(max_z, z)

        return Bounds(start=Vec3(min_x, min_y, min_z), end=Vec3(max_x, max_y, max_z))


    def get(self, x, y, z, default=0):
        x, y, z = int(x), int(y), int(z)
        try:
            return self[x][y][z]
        except:
            return default


    def reset(self):
        for x,y,z, _ in enumerate_3d(self):
            self[x][y][z] = self.default_value


    def paste(self, data, x, y, z, ignore=-1):
        for true_x in range(x, min(self.width, x+data.width)):
            for true_y in range(y, min(self.height, y+data.height)):
                for true_z in range(z, min(self.depth, y+data.depth)):
                    if data[true_x-x][true_y-y][true_z-z] == ignore:
                        continue
                    self[true_x][true_y][true_z] = data[true_x-x][true_y-y][true_z-z]


def chunk_list(target_list, chunk_size):
    # yield successive chunks from list
    for i in range(0, len(target_list), chunk_size):
        yield target_list[i:i + chunk_size]


def rotate_2d_list(target_2d_list):
    return [list(row) for row in zip(*target_2d_list[::-1], strict=True)]  # rotate


def rotate_3d_list(target_3d_list, clockwise=True): # rotates around the y (up) axis where 1 is 90 degrees clockwise
    new_data = Array3D(width=target_3d_list.depth, height=target_3d_list.height, depth=target_3d_list.width)
    print('rotate array3d, clockwise:', clockwise)

    if clockwise:
        for (x, y, z), value in enumerate_3d(target_3d_list):
            new_x = z
            new_z = target_3d_list.width - 1 - x
            new_data[new_x][y][new_z] = value
    else:
        for (x, y, z), value in enumerate_3d(target_3d_list):
            new_x = target_3d_list.depth - 1 - z
            new_z = x
            new_data[new_x][y][new_z] = value

    return new_data


def string_to_2d_list(string, char_value_map=None): # char_value_map default: {'.': 0, '#': 1}
    from textwrap import dedent
    if char_value_map is None:
        char_value_map = {'.': 0, '#': 1}
    grid = dedent(string).strip()
    grid = grid.split('\n')
    grid = [[char_value_map.get(e, 0) for e in line] for line in grid]
    grid = [list(row) for row in zip(*grid[::-1], strict=True)]  # rotate
    return grid

if __name__ == '__main__':
    from ursina.ursinastuff import _test

    _test(string_to_2d_list('''\
        #..#.###..####..#..
        #..#.#..#.###...#..
        #..#.###.....#..#..
        .##..#..#.####..#..
        ''')
        ==
        rotate_2d_list([
        [1,0,0,1, 0, 1,1,1,0, 0, 1,1,1,1, 0, 0,1,0,0],
        [1,0,0,1, 0, 1,0,0,1, 0, 1,1,1,0, 0, 0,1,0,0],
        [1,0,0,1, 0, 1,1,1,0, 0, 0,0,0,1, 0, 0,1,0,0],
        [0,1,1,0, 0, 1,0,0,1, 0, 1,1,1,1, 0, 0,1,0,0],
    ])
    )


def list_2d_to_string(target_2d_list, characters='.#'):
    return '\n'.join([''.join([characters[e] for e in line]) for line in target_2d_list])

if __name__ == '__main__':
    list_2d = [
        [1,0,0,1, 0, 1,1,1,0, 0, 1,1,1,1, 0, 0,1,0,0],
        [1,0,0,1, 0, 1,0,0,1, 0, 1,1,1,0, 0, 0,1,0,0],
        [1,0,0,1, 0, 1,1,1,0, 0, 0,0,0,1, 0, 0,1,0,0],
        [0,1,1,0, 0, 1,0,0,1, 0, 1,1,1,1, 0, 0,1,0,0],
    ]
    from textwrap import dedent
    expected_result = dedent('''
        #..#.###..####..#..
        #..#.#..#.###...#..
        #..#.###.....#..#..
        .##..#..#.####..#..
        ''').strip()

    _test(list_2d_to_string(list_2d) == expected_result)


def sample_bilinear(target_2d_list:Array2D, x, y, clamp_if_outside=True):
    X = clamp(int(x), 0, target_2d_list.width-1)
    Y = clamp(int(y), 0, target_2d_list.height-1)
    next_x = clamp(X+1, 0, target_2d_list.width-1)
    next_y = clamp(Y+1, 0, target_2d_list.height-1)

    height_bottom_left = target_2d_list[X][Y]
    height_bottom_right = target_2d_list[next_x][Y]
    height_top_left = target_2d_list[X][next_y]
    height_top_right = target_2d_list[next_x][next_y]

    x_fraction = x - X
    y_fraction = y - Y

    bottom_row_interpolation = (1 - x_fraction) * height_bottom_left + x_fraction * height_bottom_right
    top_row_interpolation = (1 - x_fraction) * height_top_left + x_fraction * height_top_right
    interpolated_height = (1 - y_fraction) * bottom_row_interpolation + y_fraction * top_row_interpolation

    return interpolated_height

if __name__ == '__main__':
    list_2d = Array2D(data=[
        [1,1,0],
        [1,0,0],
        [1,0,0],
    ])
    _test(sample_bilinear(list_2d, 0, 0) == 1)
    _test(sample_bilinear(list_2d, 1, 1) == 0)
    _test(sample_bilinear(list_2d, .5, .5) == .75)
    _test(sample_bilinear(list_2d, .5, .75) == 0.625)



class LoopingList(list):
    def __getitem__(self, i):
        return super().__getitem__(i % len(self))




from ursina import *


app = Ursina(size=(1280,720))

camera.orthographic = True  # remove perspective
camera.fov = 720            # set the height of the camera to be 720 units

Sprite.ppu = 1  # pixels per unit
player = Sprite('ursina_wink_0000')

def update():
    player.x += int(held_keys['d'] * 1)     # move 10 pixels per second


app.run()



import sys
import os

from ursina import application
from ursina.scene import instance as scene
from ursina.sequence import Sequence, Func, Wait


class Empty():
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key ,value)

class DotDict(dict):
    """Custom dictionary class to allow dot notation access."""
    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(f"'DotDict' object has no attribute '{attr}'")

    def __setattr__(self, attr, value):
        self[attr] = value

    def __delattr__(self, attr):
        try:
            del self[attr]
        except KeyError:
            raise AttributeError(f"'DotDict' object has no attribute '{attr}'")


class Default:
    pass


def invoke(function, *args, **kwargs):  # reserved keywords: 'delay', 'unscaled'
    delay = 0
    if 'delay' in kwargs:
        delay = kwargs['delay']
        del kwargs['delay']

    unscaled = False
    if 'unscaled' in kwargs:
        unscaled = kwargs['unscaled']
        del kwargs['unscaled']

    ignore_paused = False
    if 'ignore_paused' in kwargs:
        ignore_paused = kwargs['ignore_paused']
        del kwargs['ignore_paused']

    if not delay:
        function(*args, **kwargs)
        return None

    if ignore_paused:
        unscaled = True

    return Sequence(
        Wait(delay),
        Func(function, *args, **kwargs),
        auto_destroy=True, ignore_paused=ignore_paused, unscaled=unscaled, started=True,
    )


def after(delay, unscaled=True, ignore_paused=False, entity=None):    # function for @after decorator. Use the decorator, not this.
    '''@after decorator for calling a function after some time.

        example:
        @after(.4)
        def reset_cooldown():
            self.on_cooldown = False
            self.color = color.green
    '''
    def _decorator(func):
        def wrapper(*args, **kwargs):
            sequence = invoke(func, *args, **kwargs, delay=delay, unscaled=unscaled, ignore_paused=ignore_paused)
            if entity is not None:
                entity.animations.append(sequence)

        return wrapper()
    return _decorator


def size_list():    # return a list of current python objects sorted by size
    import operator

    globals_list = []
    globals_list.clear()
    for e in globals():
        # object, size
        globals_list.append([e, sys.getsizeof(e)])
    globals_list.sort(key=operator.itemgetter(1), reverse=True)
    print('scene size:', globals_list)


def find_sequence(name, file_types, folders): # find frame_0, frame_1, frame_2 and so on
    for folder in folders:
        for file_type in file_types:
            files = list(folder.glob(f'**/{name}*.{file_type}'))
            if files:
                files.sort()
                return files
    return []


def import_all_classes(path=application.asset_folder, debug=False):
    path = str(path)
    sys.path.append(path)
    from ursina.string_utilities import snake_to_camel
    from glob import iglob
    imported_successfully = []

    for file_path in iglob(path + '**/*.py', recursive=True):
        if '\\build\\' in file_path or '__' in file_path:
            continue

        rel_path = file_path[len(path):][:-3].replace('\\', '.')
        if rel_path.startswith('.'):
            rel_path = rel_path[1:]
        module_name = os.path.basename(file_path).split('.')[0]
        # class_name = snake_to_camel(module_name)
        module_name = module_name
        import_statement = 'from ' + rel_path + ' import *'

        try:
            exec(import_statement, globals())
            imported_successfully.append(module_name)
            if debug:
                print(import_statement)
        except:
            if debug:
                print('     x', import_statement)
            pass

    return imported_successfully


def print_on_screen(text, position=(0,0), origin=(-.5,.5), scale=1, duration=1, color=(1,1,1,1)):
    from ursina.text import Text
    from ursina import destroy
    text_entity = Text(text=text, position=position, origin=origin, scale=scale, color=color)
    destroy(text_entity, delay=duration)


import traceback
from inspect import getframeinfo, stack
import ast
import textwrap
import inspect

def _test(result):
    if callable(result):
        result = result()

    caller_frame = stack()[1][0]
    caller_info = getframeinfo(caller_frame)
    filename, lineno = caller_info.filename, caller_info.lineno

    # Get full multi-line _test call source
    source_lines, starting_line_no = inspect.getsourcelines(caller_frame)
    rel_lineno = lineno - starting_line_no
    lines = source_lines[rel_lineno:]
    call_text = ""
    parens = 0
    started = False
    for line in lines:
        stripped = line.strip()
        if not started and stripped.startswith("_test"):
            started = True
        if started:
            call_text += line
            parens += line.count("(") - line.count(")")
            if parens <= 0:
                break
    call_text = textwrap.dedent(call_text)

    try:
        tree = ast.parse(call_text, mode='exec')
        call = tree.body[0].value  # _test(...)
        arg_source = ast.unparse(call.args[0])
    except Exception:
        arg_source = "<expression>"

    def extract_func_name(expr_ast):
        if isinstance(expr_ast, ast.Call):
            return extract_func_name(expr_ast.func)
        elif isinstance(expr_ast, ast.Attribute):
            value = extract_func_name(expr_ast.value)
            return f"{value}.{expr_ast.attr}"
        elif isinstance(expr_ast, ast.Name):
            return expr_ast.id
        else:
            return "<unknown>"

    try:
        expr_ast = ast.parse(arg_source, mode='eval').body
        if isinstance(expr_ast, ast.Compare):
            func_name = extract_func_name(expr_ast.left)
        else:
            func_name = extract_func_name(expr_ast)
    except Exception:
        func_name = "<unknown>"

    MAX_LEN = 32
    display_expr = (arg_source[:MAX_LEN] + '...') if len(arg_source) > MAX_LEN else arg_source

    if result:
        print(f'\33[42mPASSED\033[0m {func_name} {display_expr}')
        return True
    else:
        print(f'\33[41mFAILED\033[0m {func_name} {display_expr}')
        print(f' --> {filename}:{lineno}')
        print("Stack trace (most recent call last):")
        traceback.print_stack(f=caller_frame)
        return False




# define a new metaclass which overrides the "__call__" function
class PostInitCaller(type):
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj.__post_init__()
        return obj

if __name__ == '__main__':

    from ursina import *
    app = Ursina()





    a = Audio('sine')
    a.play()
    destroy(a, delay=1)
    # Player()
    app.run()



from ursina import *


class Draggable(Button):
    _z_plane = None

    def __init__(self, require_key=None, step=Vec3.zero, plane_direction=Vec3.forward, lock=None, min_x=-inf, min_y=-inf, min_z=-inf, max_x=inf, max_y=inf, max_z=inf, **kwargs):
        if not __class__._z_plane:
            __class__._z_plane = Entity(name='_z_plane', scale=(9999,9999), enabled=False, eternal=True)

        super().__init__(**kwargs)
        self.require_key = require_key
        self.dragging = False
        self.delta_drag = 0
        self.start_pos = self.world_position
        self.start_offset = (0,0,0)
        self.step = step
        self.plane_direction = plane_direction
        self.lock = lock if lock else Vec3(0,0,0)     # set to 1 to lock movement on any of x, y and z axes
        self.min_x = min_x
        self.min_y = min_y
        self.min_z = min_z
        self.max_x = max_x
        self.max_y = max_y
        self.max_z = max_z

        if not Draggable._z_plane.model: # set these after game start so it can load the model
            Draggable._z_plane.model = 'quad'
            Draggable._z_plane.collider = Mesh(vertices=((-0.5, -0.5, 0.0), (0.5, -0.5, 0.0), (0.5, 0.5, 0.0), (-0.5, 0.5, 0.0)), triangles=((0,1,2,3),), mode='triangle')
            Draggable._z_plane.color = color.clear


        for key, value in kwargs.items():
            if key == 'collider' and value == 'sphere' and self.has_ancestor(camera.ui):
                raise Exception('error: sphere colliders are not supported on Draggables in ui space.')

            if key == 'text' or key in self.attributes:
                continue

            setattr(self, key, value)


    def input(self, key):
        if self.hovered and key == 'left mouse down':
            if self.require_key == None or held_keys[self.require_key]:
                self.start_dragging()

        if self.dragging and key == 'left mouse up':
            self.stop_dragging()


    def start_dragging(self):
        point = Vec3(0,0,0)
        if mouse.world_point:
            point = mouse.world_point

        Draggable._z_plane.world_position = point
        # Draggable._z_plane.world_position = self.world_position
        Draggable._z_plane.look_at(Draggable._z_plane.position - Vec3(*self.plane_direction))
        if self.has_ancestor(camera.ui):
            Draggable._z_plane.world_parent = camera.ui
        else:
            Draggable._z_plane.world_parent = scene

        self.start_offset = point - self.world_position
        self.dragging = True
        self.start_pos = self.world_position
        self.collision = False
        Draggable._z_plane.enabled = True
        mouse._original_traverse_target = mouse.traverse_target
        mouse.traverse_target = Draggable._z_plane
        if hasattr(self, 'drag'):
            self.drag()


    def stop_dragging(self):
        self.dragging = False
        self.delta_drag = self.world_position - self.start_pos
        Draggable._z_plane.enabled = False
        self.collision = True
        if hasattr(mouse, '_original_traverse_target'):
            mouse.traverse_target = mouse._original_traverse_target
        else:
            mouse.traverse_target = scene

        if hasattr(self, 'drop'):
            self.drop()

    # def drag(self):
    #     print('start drag test')
    #
    # def drop(self):
    #     print('drop test')

    def update(self):
        if self.dragging:
            if mouse.world_point:
                if not self.lock[0]:
                    self.world_x = mouse.world_point[0] - self.start_offset[0]
                if not self.lock[1]:
                    self.world_y = mouse.world_point[1] - self.start_offset[1]
                if not self.lock[2]:
                    self.world_z = mouse.world_point[2] - self.start_offset[2]

            if self.step[0] > 0:
                hor_step = 1/self.step[0]
                self.x = round(self.x * hor_step) /hor_step
            if self.step[1] > 0:
                ver_step = 1/self.step[1]
                self.y = round(self.y * ver_step) /ver_step
            if self.step[2] > 0:
                dep_step = 1/self.step[2]
                self.z = round(self.z * dep_step) /dep_step

        self.position = (
            clamp(self.x, self.min_x, self.max_x),
            clamp(self.y, self.min_y, self.max_y),
            clamp(self.z, self.min_z, self.max_z)
            )


    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, value):
        if isinstance(value, int | float | complex):
            value = (value, value, value)

        self._step = value



if __name__ == '__main__':
    app = Ursina()

    Entity(model='plane', scale=8, texture='white_cube', texture_scale=(8,8))
    draggable_button = Draggable(scale=.1, text='drag me', position=(-.5, 0))
    world_space_draggable = Draggable(parent=scene, model='cube', color=color.azure, plane_direction=(0,1,0), lock=(1,0,0))

    EditorCamera(rotation=(30,10,0))
    world_space_draggable.drop = Func(print, 'dropped cube')

    app.run()



from ursina import color
from ursina.shader import Shader
from ursina.vec3 import Vec3

fog_of_war_shader = Shader(language=Shader.GLSL,
vertex = '''
#version 140
uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelMatrix;

in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 texcoord;

in vec3 p3d_Normal;
out vec3 world_normal;
uniform vec3 light_position;
out float dist;

void main() {
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    texcoord = p3d_MultiTexCoord0;

    // dist = length(light_position - p3d_Vertex.rgb) * .1;
    dist = length(light_position - gl_Position.xyz) * .01;
    // dist = 1.0;
}
''',

fragment='''
#version 140

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
in vec2 texcoord;
in vec3 world_normal;
out vec4 fragColor;
in float dist;

// uniform mat4 p3d_ViewProjectionMatrix;
// uniform sampler2D tex;
// uniform sampler2D dtex;
// in vec2 uv;
// out vec4 o_color;

// uniform vec3 light_position;

// vec3 reconstructPosition(in vec2 uv, in float z)
// {
//     float x = uv.x * 2.0f - 1.0f;
//     float y = (1.0 - uv.y) * 2.0f - 1.0f;
//     vec4 position_s = vec4(x, y, z, 1.0f);
//     mat4x4 view_projection_matrix_inverse = inverse(p3d_ViewProjectionMatrix);
//     vec4 position_v = view_projection_matrix_inverse * position_s;
//     return position_v.xyz / position_v.w;
// }

void main() {
    vec4 color = texture(p3d_Texture0, texcoord) * p3d_ColorScale * (1-dist);
    color.a = 1.;
    fragColor = color.rgba;


    // float depth = texture(dtex, uv).r;
    // vec3 position = reconstructPosition(uv, depth);
    //
    // float dist = (position - light_position).x;
    //
    // o_color.rgb = texture(tex, uv).rgb * dist;
    // // o_color.rgb = vec3(dist);
    // o_color.a = 1.0;
}


''', geometry='',
default_input = {
    'light_position' : Vec3(0,1,0),

}
)


if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    window.color=color.black

    Entity.default_shader = fog_of_war_shader

    editor_camera = EditorCamera()
    # camera.shader = fog_of_war_shader

    ground = Entity(model='plane', collider='box', scale=64, texture='grass', texture_scale=(4,4))
    for i in range(16):
        Entity(model='cube', origin_y=-.5, scale=2, texture='brick', texture_scale=(1,2),
            x=random.uniform(-8,8),
            z=random.uniform(-8,8) + 8,
            collider='box',
            scale_y = random.uniform(2,3),
            color=color.hsv(0, 0, random.uniform(.9, 1))
            )

    light = Entity(model='sphere', unlit=True, scale=.5, y=1)

    def update():
        light.x += held_keys['d'] - held_keys['a']
        light.z += held_keys['w'] - held_keys['s']

        # camera.set_shader_input('light_position', light.world_position)
        for e in scene.entities:
            if hasattr(e,'shader') and e.shader == fog_of_war_shader:
                e.set_shader_input('light_position', light.world_position + camera.world_position)
                # print(e)


    app.run()
